# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Device Power Default Capability."""

import time
from typing import Any, Callable, Dict, Optional

from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.capabilities.interfaces import device_power_base
from gazoo_device.utility import deprecation_utils
import immutabledict

logger = gdm_logger.get_logger()

_HUB_TYPE_PROPERTY = "device_power_hub_type"
_HUB_TYPE_CAMBRIONIX = "cambrionix"
_HUB_TYPE_POWERSWITCH = "powerswitch"
_HUB_TYPE_UNIFI_SWITCH = "unifi_switch"

SUPPORTED_HUB_TYPES = (
    _HUB_TYPE_CAMBRIONIX,
    _HUB_TYPE_POWERSWITCH,
    _HUB_TYPE_UNIFI_SWITCH,
)

_HUB_TYPE_PROPS = immutabledict.immutabledict({
    _HUB_TYPE_CAMBRIONIX: ("device_usb_hub_name", "device_usb_port"),
    _HUB_TYPE_POWERSWITCH: ("powerswitch_name", "powerswitch_port"),
    _HUB_TYPE_UNIFI_SWITCH: ("unifi_switch_name", "unifi_switch_port"),
})


class DevicePowerDefault(device_power_base.DevicePowerBase):
  """Base class for device_power."""

  def __init__(
      self,
      device_name: str,
      create_device_func: Callable[..., Any],
      default_hub_type: str,
      props: Dict[str, Any],
      usb_ports_discovered: bool,
      wait_for_bootup_complete_fn: Callable[[], None],
      switchboard_inst: Any,
      change_triggers_reboot: bool = False,
      usb_hub_name_prop: str = "device_usb_hub_name",
      usb_port_prop: str = "device_usb_port",
      wait_for_connection_fn: Optional[Callable[[], None]] = None,
  ):
    """Create an instance of the device_power capability.

    The power switch type can be changed by setting the 'device_power_hub_type'
    property. The valid hub types are 'cambrionix', 'powerswitch', and
    'unifi_switch'.

    Each hub type also requires a corresponding hub_name and hub_port property
    to be set.

    Cambrionix requires 'device_usb_hub_name' and 'device_usb_port' to be set.
      (These names can be changed by passing in the different property names)

    Powerswitch requires 'powerswitch_name' and 'powerswitch_port' to be set.

    Unifi_switch requires 'unifi_switch_name' and 'unifi_switch_port' to be set.

    Args:
      device_name: Name of the device this capability is attached to.
      create_device_func: A method to create device of hub_type.
      default_hub_type: Type of switch for power cycling.
      props: Dictionary of device props from configuration file.
      usb_ports_discovered: True if the USB ports are discovered by gdm detect.
      wait_for_bootup_complete_fn: A method that the capability can call to wait
        for a reboot to complete if triggered by a change.
      switchboard_inst: Instance of switchboard capability
      change_triggers_reboot: Set change_triggers_reboot to TRUE if changing the
        power mode for the device causes a reboot.
      usb_hub_name_prop: Name of the hub name property.
      usb_port_prop: Name of the hub port property.
      wait_for_connection_fn: a method to wait for the device to become
        reachable.

    Raises:
      ValueError: If hub type is not a known type.
    """
    super().__init__(device_name=device_name)

    self._hub = None
    self._create_device_func = create_device_func
    self._usb_ports_discovered = usb_ports_discovered
    self._props = props
    hub_type = self._props["optional"].get(_HUB_TYPE_PROPERTY, default_hub_type)
    if hub_type not in SUPPORTED_HUB_TYPES:
      raise ValueError("Hub type {} is not supported. Valid types: {}".format(
          hub_type, SUPPORTED_HUB_TYPES))
    self._hub_type = hub_type
    self._hub_name_prop, self._port_prop = _HUB_TYPE_PROPS[self._hub_type]
    if self._hub_type == _HUB_TYPE_CAMBRIONIX:
      self._hub_name_prop = usb_hub_name_prop
      self._port_prop = usb_port_prop

    # Set where the hub/port propeties are in _props.
    if self._hub_type == _HUB_TYPE_CAMBRIONIX and self._usb_ports_discovered:
      self._dict_name = "persistent_identifiers"
    else:
      self._dict_name = "optional"

    self._wait_for_bootup_complete_fn = wait_for_bootup_complete_fn
    self._switchboard = switchboard_inst
    self._change_triggers_reboot = change_triggers_reboot
    self._wait_for_connection_fn = wait_for_connection_fn

  @decorators.CapabilityLogDecorator(logger, level=decorators.DEBUG)
  def health_check(self):
    """Checks that the capability is ready to use.

    Raises:
      CapabilityNotReadyError: If expected properties are not set or
        unable to create auxiliary device for power switching.
    """
    unset_props = []
    if self.hub_name is None:
      unset_props.append(self._hub_name_prop)
    if self.port_number is None:
      unset_props.append(self._port_prop)
    if unset_props:
      if self._hub_type == _HUB_TYPE_CAMBRIONIX and self._usb_ports_discovered:
        msg_format = ("If device is connected to {}, "
                      "set them via 'gdm redetect {}'")
      else:
        msg_format = ("If device is connected to {}, "
                      "set them via 'gdm set-prop {} <property> <value>'")
      msg = msg_format.format(self._hub_type, self._device_name)
      error_msg = "properties {} are unset. ".format(
          " and ".join(unset_props)) + msg
      raise errors.CapabilityNotReadyError(
          msg=error_msg, device_name=self._device_name)
    try:
      self._hub = self._create_device_func(self.hub_name)
    except errors.DeviceError as err:
      raise errors.CapabilityNotReadyError(
          msg=str(err), device_name=self._device_name)
    self._healthy = True

  @decorators.CapabilityLogDecorator(logger, level=decorators.DEBUG)
  def close(self):
    """Closes the hub device instance."""
    if self._hub:
      self._hub.close()
    super().close()

  @decorators.OptionalProperty
  def hub_name(self):
    """Name of the hub the device is attached to."""
    return self._props[self._dict_name].get(self._hub_name_prop)

  @decorators.OptionalProperty
  def hub_type(self):
    """Device type of the hub."""
    return self._hub_type

  @decorators.DynamicProperty
  def port_mode(self):
    """port mode."""
    if not self.healthy:
      self.health_check()
    return self._hub.switch_power.get_mode(self.port_number)

  @decorators.OptionalProperty
  def port_number(self):
    """Port number the device is attached to."""
    return self._props[self._dict_name].get(self._port_prop)

  @decorators.CapabilityLogDecorator(logger)
  def cycle(self, no_wait=False):
    """Power off then power on the device.

    Args:
        no_wait (bool):  Return before verifying boot up.
    """
    if not self.healthy:
      self.health_check()
    self.off()
    time.sleep(2)  # Small delay before calling power_on
    self.on(no_wait=no_wait)

  @decorators.CapabilityLogDecorator(logger)
  def off(self, close_transports=True):
    """Turn off power to the device.

    Args:
      close_transports (bool): Whether to close transports when powering off.
    """
    if not self.healthy:
      self.health_check()
    # No-op if port is already off.
    status = self._hub.switch_power.get_mode(self.port_number)
    if status == "off":
      return
    if self._change_triggers_reboot:
      self._switchboard.add_log_note(
          f"GDM triggered reboot via {self.hub_type} power change.")
      self._hub.switch_power.power_off(self.port_number)
      self._wait_for_bootup_complete_fn()
    else:
      if close_transports:
        self._switchboard.close_all_transports()
      self._hub.switch_power.power_off(self.port_number)

  @decorators.CapabilityLogDecorator(logger)
  def on(self, no_wait=False):
    """Turn on power to the device.

    If Switchboard hasn't been initialized then transports won't be opened
    and GDM triggered reboot won't be logged to avoid initializing the
    switchboard. Note that if no_wait=False then Switchboard will be initialized
    at the end.

    Args:
        no_wait (bool):  Return before verifying boot up.
    """
    if not self.healthy:
      self.health_check()
    # No-op if port is already on.
    status = self._hub.switch_power.get_mode(self.port_number)
    if status in ["sync", "charge", "on"]:
      return
    if self._change_triggers_reboot:
      if self._switchboard.health_checked:
        self._switchboard.add_log_note(
            f"GDM triggered reboot via {self.hub_type} power change.")
    self._hub.switch_power.power_on(self.port_number)
    if not self._change_triggers_reboot:
      if self._wait_for_connection_fn is not None:
        self._wait_for_connection_fn()
      else:
        time.sleep(2)  # Small delay to give time for 'dev/tty' to populate
      if self._switchboard.health_checked:
        self._switchboard.open_all_transports()

    if not no_wait:
      self._wait_for_bootup_complete_fn()


deprecation_utils.add_deprecated_attributes(DevicePowerDefault,
                                            [("power_off", "off", True),
                                             ("power_on", "on", True),
                                             ("power_cycle", "cycle", True)])



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for interacting with devices in parallel.

Usage example:

  def my_custom_hello_world_function(manager_inst: manager.Manager,
                                     device_name: str, some_arg: int) -> str:
    '''Example function which executes a custom action on a device.

    Args:
      manager_inst: A Manager instance which can be used for device creation.
        A Manager instance is always provided. You do not need to pass one as an
        argument to the function in the CallSpec.
      device_name: Name of the device to use. Must be specified in the CallSpec.
      some_arg: An example of an argument. Must be specified in the CallSpec.
    '''
    device = manager_inst.create_device(device_name)
    try:
      shell_response = device.shell(f"echo 'Hello world {some_arg}'")
    finally:
      device.close()
    return shell_response

  call_specs = [
      parallel_utils.CallSpec(parallel_utils.reboot, "device-1234"),
      parallel_utils.CallSpec(parallel_utils.reboot, "device-2345",
                              no_wait=True, method="shell"),
      parallel_utils.CallSpec(parallel_utils.factory_reset, "device-3456"),
      parallel_utils.CallSpec(parallel_utils.upgrade, "device-4567",
                              build_number=1234, build_branch="1.0"),
      parallel_utils.CallSpec(parallel_utils.upgrade, "device-5678",
                              build_file="/some/file/path.zip"),
      parallel_utils.CallSpec(my_custom_hello_world_function, "device-6789",
                              1),
      parallel_utils.CallSpec(my_custom_hello_world_function, "device-7890",
                              some_arg=2),
  ]

  results, _ = parallel_utils.execute_concurrently(
      call_specs, timeout=300, raise_on_process_error=True)

Results are returned in the same order as call specs. For the hypothetical
example above, results would be
  [None, None, None, None, None, "Hello world 1", "Hello world 2"].

If you need more granular control over exceptions raised in parallel processes,
set raise_on_process_error to False. For example:

  def custom_function_raises(
      manager_inst: manager.Manager, some_arg: int) -> NoReturn:
    raise RuntimeError(f"Demo of exception handling {some_arg}")

  results, errors = parallel_utils.execute_concurrently(
      call_specs = [
          parallel_utils.CallSpec(custom_function_raises, 1),
          parallel_utils.CallSpec(custom_function_raises, 2),
      ],
      timeout=15,
      raise_on_process_error=False)

In this case results will be ["< No result received >",
                              "< No result received >"].
Errors will be [
    ("RuntimeError", "Demo of exception handling 1", "< Exception traceback >"),
    ("RuntimeError", "Demo of exception handling 2", "< Exception traceback >"),
].

Logging behavior:
  Parallel process GDM logger logs are sent to the main process.
  Device logs (from device instances created in parallel processes) are stored
  in new individual device log files.
"""
import concurrent.futures
import dataclasses
import importlib
import multiprocessing
import os
import time
import traceback
from typing import Any, Callable, List, Optional, Sequence, Tuple

from gazoo_device import errors
from gazoo_device import extensions
from gazoo_device import gdm_logger
from gazoo_device import manager
from gazoo_device import package_registrar
from gazoo_device.utility import multiprocessing_utils
import immutabledict

NO_RESULT = "< No result received >"
NO_TRACEBACK = "< No traceback available >"
TIMEOUT_PROCESS = 600.0
_TIMEOUT_TERMINATE_PROCESS = 3
_QUEUE_READ_TIMEOUT = 1
_AnySerializable = Any


@dataclasses.dataclass(init=False)
class CallSpec:
  """Specifies a call to be executed in a parallel process.

  The function will be called in a parallel process as follows:
    return_value = function(<ManagerInstance>, *args, **kwargs)

  A Manager instance is always provided as the first argument to the function,
  followed by *args and **kwargs. The Manager instance will be closed
  automatically after the function returns.

  If the function is performing a device action, it is expected to create a
  device instance using the provided Manager instance (the device name to create
  should be included in the function's arguments), use the device instance to
  perform some action (possibly parameterized by args and kwargs), and close the
  device instance before returning. The return_value of the function will be
  returned to the main process.

  Attributes:
    function: Function to call in the parallel process. The function and its
      return value must be serializable. Prefer module-level functions. In
      particular, lambdas and inner (nested) functions are not serializable.
      Other limitations:
      - For devices communicating over UART or serial: ensure that access to
        device communication is mutually exclusive. In particular, make sure
        that device communication (`<device>.switchboard`) is closed in the main
        process before issuing a parallel action on it and do not execute
        simultaneous parallel actions on the same device.
        `<device>.reset_capability("switchboard")` can be used to close device
        communication, and it will be automatically reopened on the next access.
        - Do not modify GDM device configs (detection, set-prop) in parallel.
        This can result in a race condition.
    args: Positional arguments to the function. Must be serializable. In
      particular, Manager and device instances as well their instance methods
      are not serializable.
    kwargs: Keyword arguments to the function. Must be serializable.
  """
  function: Callable[..., _AnySerializable]
  args: Tuple[_AnySerializable, ...]
  kwargs: immutabledict.immutabledict[str, _AnySerializable]

  def __init__(self, function: Callable[..., _AnySerializable],
               *args: _AnySerializable, **kwargs: _AnySerializable):
    self.function = function
    self.args = args
    self.kwargs = immutabledict.immutabledict(kwargs)


def _process_init(
    logging_queue: multiprocessing.Queue,
    extension_package_import_paths: Sequence[str],
) -> None:
  """Initializes worker process."""
  gdm_logger.initialize_child_process_logging(logging_queue)
  logger = gdm_logger.get_logger()
  short_description = f"_process_init in process {os.getpid()}"
  # The state of the main process (such as registered extension packages) is not
  # copied over when using "forkserver" or "spawn" as the process start method.
  for import_path in extension_package_import_paths:
    try:
      package_registrar.register(importlib.import_module(import_path))
    except (ImportError, errors.PackageRegistrationError) as e:
      logger.debug(f"{short_description}: failed to import and register GDM "
                   f"extension package with import path {import_path}. "
                   f"Error: {e!r}. Proceeding despite the failure.")


def _process_wrapper(call_spec: CallSpec) -> Any:
  """Executes the provided function in a parallel process."""
  logger = gdm_logger.get_logger()
  short_description = f"{call_spec.function.__name__} in process {os.getpid()}"
  logger.debug(f"{short_description}: starting execution of {call_spec}...")

  manager_inst = manager.Manager()
  try:
    return_value = call_spec.function(manager_inst, *call_spec.args,
                                      **call_spec.kwargs)
    logger.debug(f"{short_description}: execution succeeded. "
                 f"Return value: {return_value}.")
    return return_value
  finally:
    manager_inst.close()


def _format_process_errors(
    proc_errors: Sequence[Optional[Tuple[str, str, str]]]) -> str:
  """Returns a formatted string with all process errors."""
  formatted_errors = []
  for proc_error in proc_errors:
    if proc_error is not None:
      error_type, error_message, error_tb = proc_error
      if error_tb != NO_TRACEBACK:
        formatted_error = error_tb
      else:
        formatted_error = f"{error_type}({error_message!r})"
      formatted_errors.append(formatted_error)
  return "\n".join(formatted_errors)


def execute_concurrently(
    call_specs: Sequence[CallSpec],
    timeout: float = TIMEOUT_PROCESS,
    raise_on_process_error: bool = True,
    max_processes: Optional[int] = None,
) -> Tuple[List[Any], List[Optional[Tuple[str, str, str]]]]:
  """Concurrently executes function calls in parallel processes.

  Args:
    call_specs: Specifications for each of the parallel executions.
    timeout: Time to wait before terminating all of the parallel processes.
    raise_on_process_error: If True, raise an error if any of the parallel
      processes encounters an error. If False, return a list of errors which
      occurred in the parallel processes along with the received results.
    max_processes: Maximum number of processes to use for the tasks. If set to
      None then os.cpu_count() is used. See docs for futures.ProcessPoolExecutor
      https://docs.python.org/3/library/concurrent.futures.html#concurrent.futures.ProcessPoolExecutor

  Returns:
    A tuple of (parallel_process_return_values, parallel_process_errors).
    The order of return values and errors corresponds to the order of provided
    call_specs. parallel_process_return_values will contain return values of the
    functions executed in parallel. If a parallel process fails, the
    corresponding entry in the return value list will be NO_RESULT.
    Errors are only returned if raise_on_process_error is False.
    Each error is specified as a tuple of
    (error_type, error_message, error_traceback). If a parallel process succeeds
    (there's no error), the corresponding entry in the error list will be None.

  Raises:
    ParallelUtilsError: If raise_on_process_error is True and any of the
      parallel processes encounters an error.
  """
  gdm_logger.switch_to_multiprocess_logging()
  logging_queue = gdm_logger.get_logging_queue()
  extension_package_import_paths = [
      package_info["import_path"]
      for package_name, package_info in extensions.package_info.items()
      if package_name != "gazoo_device_controllers"  # Built-in controllers.
  ]

  proc_results = []
  proc_errors = []
  futures = []
  with concurrent.futures.ProcessPoolExecutor(
      max_workers=max_processes,
      mp_context=multiprocessing_utils.get_context(),
      initializer=_process_init,
      initargs=(logging_queue, extension_package_import_paths)) as executor:
    for call_spec in call_specs:
      futures.append(executor.submit(_process_wrapper, call_spec=call_spec))

    deadline = time.time() + timeout
    # Get results and errors.
    for future in futures:
      try:
        remaining_timeout = max(0, deadline - time.time())
        proc_results.append(future.result(timeout=remaining_timeout))
        proc_errors.append(None)
      except concurrent.futures.TimeoutError as e:
        future.cancel()
        proc_results.append(NO_RESULT)
        proc_errors.append((
            errors.ResultNotReceivedError.__name__,
            "Did not receive any results from the process.",
            NO_TRACEBACK))
      except Exception as e:  # pylint: disable=broad-except
        proc_results.append(NO_RESULT)
        proc_errors.append((type(e).__name__, str(e), traceback.format_exc()))

  if raise_on_process_error and any(proc_errors):
    raise errors.ParallelUtilsError(
        "Encountered errors in parallel processes:\n"
        f"{_format_process_errors(proc_errors)}")

  return proc_results, proc_errors


def factory_reset(manager_inst: manager.Manager, device_name: str) -> None:
  """Convenience function for factory resetting devices in parallel."""
  device = manager_inst.create_device(device_name)
  try:
    device.factory_reset()
  finally:
    device.close()


def reboot(manager_inst: manager.Manager, device_name: str, *reboot_args: Any,
           **reboot_kwargs: Any) -> None:
  """Convenience function for rebooting devices in parallel."""
  device = manager_inst.create_device(device_name)
  try:
    device.reboot(*reboot_args, **reboot_kwargs)
  finally:
    device.close()


def upgrade(manager_inst: manager.Manager, device_name: str, *upgrade_args: Any,
            **upgrade_kwargs: Any) -> None:
  """Convenience function for upgrading devices in parallel."""
  device = manager_inst.create_device(device_name)
  try:
    device.flash_build.upgrade(*upgrade_args, **upgrade_kwargs)
  finally:
    device.close()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Fire Manager module.

Used for CLI-specific commands and flags.
Built to work with Python Fire: https://github.com/google/python-fire.
"""
import codecs
import enum
import fnmatch
import inspect
import json
import logging
import os.path
import pydoc
import re
import sys
import textwrap
import time
from typing import Any, Collection, List, Mapping, Optional, Sequence, Tuple, Type

from gazoo_device import config
from gazoo_device import console
from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device import manager
from gazoo_device import package_registrar
from gazoo_device.utility import parallel_utils
from gazoo_device.utility import usb_utils

logger = gdm_logger.get_logger()


class AttributeClassification(enum.Enum):
  """Types of class attributes recognized by the "man" method."""
  CAPABILITY = "capability"
  DEPRECATED_METHOD = "deprecated method"
  DEPRECATED_PROPERTY = "deprecated property"
  HEALTH_CHECK = "health check"
  OTHER = "other"
  PRIVATE_METHOD = "private method"
  PROPERTY = "property"
  CLASS_PROPERTY = "class property"
  PUBLIC_METHOD = "public method"

MAX_TIME_TO_WAIT_FOR_INITATION = 5

_DOC_INDENT_SIZE = 4
# Capability attributes visible on device summary man page
# (e.g. "man cambrionix").
_VISIBLE_CAPABILITY_ATTRIBUTES = [
    AttributeClassification.PROPERTY, AttributeClassification.PUBLIC_METHOD
]
_FORMAT_DEF = "  {:22s}  {}"
_FORMAT_VAL = "\t\t\t  {}"
_FORMAT_DIC = "\t\t\t  {:25s} {!r}"


def _log_man_warning_for_multiple_flavors(
    capability_classes: Collection[Type[Any]],
    capability_name: str,
    device_type: str,
    capability_class: Type[Any]) -> None:
  """Logs 'gdm man' warning when multiple capability flavors are available.

  Args:
    capability_classes: All available capability flavors.
    capability_name: Name of the capability.
    device_type: Type of the device with this capability.
    capability_class: Capability flavor selected to print documentation for.

  Capabilities can have multiple flavors in one device class, although this is
  somewhat rare. The flavor used is determined based on device firmware at
  runtime. Since we don't know which flavor will be used without an attached
  device, log a warning and print documentation for any single flavor.
  """
  flavors = [a_cls.__name__ for a_cls in capability_classes]
  logger.warning(
      f"{len(flavors)} flavors ({flavors}) of capability {capability_name!r} "
      f"are available for {device_type}.\n"
      f"Showing documentation for flavor {capability_class}.\n")


def _run_device_action(
    manager_inst: manager.Manager, device_name: str, full_attribute_name: str,
    method_args: Tuple[Any, ...], method_kwargs: Mapping[str, Any]) -> Any:
  """Returns the result of a device method call or a property."""
  device = manager_inst.create_device(device_name)
  attribute = device
  try:
    for attribute_name in full_attribute_name.split("."):
      attribute = getattr(attribute, attribute_name)
    if callable(attribute):  # It's a method.
      return attribute(*method_args, **method_kwargs)
    else:  # It's a property.
      return attribute
  finally:
    device.close()


def pretty_print_props(props_dicts: Mapping[str, Mapping[str, Any]]):
  """Prettily logs the props dictionaries from manager.get_device_prop."""
  device_property_types = ["persistent", "optional", "dynamic"]
  manager_property_types = ["settable"]
  for property_type in device_property_types + manager_property_types:
    if property_type not in props_dicts:
      continue

    title = "{} Properties:".format(property_type.capitalize())
    logger.info("")
    logger.info(title)
    for prop_name in sorted(props_dicts[property_type]):
      prop_value = props_dicts[property_type][prop_name]
      if isinstance(prop_value, list) and len(prop_value) > 1:
        logger.info(_FORMAT_DEF.format(prop_name, ""))
        for value in prop_value:
          if callable(value):
            value = value.__name__
          logger.info(_FORMAT_VAL.format(str(value)))
        logger.info("")
      elif isinstance(prop_value, dict) and len(prop_value) > 1:
        logger.info(_FORMAT_DEF.format(prop_name, ""))
        for key in sorted(prop_value):
          logger.info(_FORMAT_DIC.format(key, prop_value[key]))
        logger.info("")

      else:
        logger.info(_FORMAT_DEF.format(prop_name, prop_value))


class FireManager(manager.Manager):
  """Manages the setup and communication of smart devices."""

  def __init__(self, debug=False, dev_debug=False, quiet=False):
    stream_debug = debug or dev_debug
    debug_level = logging.WARNING if quiet else logging.DEBUG

    super().__init__(debug_level=debug_level, stream_debug=stream_debug)

  def console(self, device_name: str, health_check: bool = False) -> None:
    """Starts an interactive device console.

    Use Tab and Shift-Tab to navigate between console windows (next/previous).

    Args:
      device_name: Name of the device to start the console for.
      health_check: Whether to run health checks before starting the console.

    Raises:
      NotImplementedError: If the device does not have a Switchboard capability.
    """
    make_device_ready = "on" if health_check else "off"
    device = self.create_device(device_name,
                                make_device_ready=make_device_ready)

    try:
      if not device.has_capabilities(["switchboard"]):
        raise NotImplementedError(
            f"{device.name} does not have a Switchboard capability, which is "
            "required for the console.")
      console_configuration = device.get_console_configuration()
      device.switchboard.set_max_log_size(0)  # Disable log rotation.
      gdm_logger.silence_progress_messages()  # To avoid interference with GUI.
      console_app = console.ConsoleApp(
          device.switchboard, device.log_file_name, console_configuration)
      console_app.run()
    finally:
      gdm_logger.reenable_progress_messages()
      device.close()

  def exec(self, identifier):
    """Alias for create_device with health checks disabled.

    Important: ensure the device is in a healthy state before using "exec"
    by running health checks via `gdm health-check device-1234`.
    Usage from the CLI: `gdm exec device-1234 - shell "echo 'foo'"`.

    Args:
      identifier (str): The identifier string (name) which specifies the device.

    Returns:
      Object: The device found and created by the identifier specified.
    """
    return self.create_device(identifier, make_device_ready="off")

  def get_persistent_prop_devices(self, devices):
    """Gets persistent props of one or more devices and returns a json response.

    This assumes the provided devices are healthy. If properties are unable to
    be fetched, an empty json object will be returned for the device for which
    the properties could not be retrieved.

    Args:
      devices (list): Device identifiers to get properties of.

    Returns:
      str: json formatted persistent properties, e.g.
        {
          'device-1234': {
            'build_date': '180124',
            'ftdi_serial_number': '1f824023'
          },
          'device-5678': {},
          ...
        }
    """
    logger.setLevel(logging.ERROR)  # silence logging to reduce CLI output
    devices_props = {}
    if isinstance(devices, str):
      devices = devices.split(",")

    for device_name in devices:
      persistent_props = {}
      try:
        device_props = self.get_device_configuration(device_name)
        persistent_props = device_props.get("persistent", {})
      except errors.DeviceError:
        pass  # unhealthy devices will have empty props
      devices_props[device_name] = persistent_props

    return json.dumps(devices_props)

  def get_prop(self, device_name, prop=None):
    """Prints the device properties.

    Args:
      device_name (str): identifier for device.
      prop (str): identifier for property.
    """
    if prop is not None:  # If requested a single property...
      value = self.get_device_prop(device_name, prop)
      logger.info(_FORMAT_DEF.format(prop, str(value)))
    else:
      props_dicts = self.get_device_prop(device_name, prop)
      pretty_print_props(props_dicts)

  def health_check(self, identifier, recover=False):
    """CLI command for running device health checks.

    Usage from the CLI: `gdm health-check device-1234`.

    Args:
      identifier (str): The identifier string (name) which specifies the
        device.
      recover (bool): whether to automatically recover from health check
        failures.
    """
    make_device_ready_setting = "check_only"
    if recover:
      make_device_ready_setting = "on"
    device = self.create_device(identifier, make_device_ready="off")
    try:
      device.make_device_ready(setting=make_device_ready_setting)
    finally:
      device.close()

  @classmethod
  def helpfull(cls):
    """Prints a general overview of GDM's features.

    Invoked by 'man' without arguments.
    """
    description = textwrap.dedent("""
      Gazoo Device Manager (gdm) command line interface is a single utility for control of a
      variety of Gazoo hardware.

      Commands:
        The CLI is dynamic, meaning commands are generated from API methods (see
        https://github.com/google/python-fire).
        There are several groups of commands.
          "Manager" commands operate on local config files on the PC.
          These commands are generated from API methods in the Manager class.
            The command may talk to devices (detect), but they don't change the state of
            the device (i.e. read-only)
            The command may change the config files stored on the PC, to detect new
            devices, add an alias, etc.
            To see a list of available manager commands, run "gdm".
          "Device" commands talk to or modify a Gazoo device.
          These commands are generated from API methods in device classes.
              This includes upgrade, reboot, etc.
              In general you enter a device command as "gdm issue <device_name> - <command>"
              Examples::
              gdm issue raspberrypi-zdwm - reboot
              gdm issue raspberrypi-zdwm - shell "echo 'foo'"
              To see the list of device commands available for a device, run "gdm man <device_type>"
              For example:: gdm man cambrionix

          You can get more details on a particular command at the command line with:
          "gdm <command> -h"
          For example::  gdm detect -h

          You can pass in flags to the CLI with:
          "gdm --<flag> - <command>"
          For example:: gdm --debug - devices
          To see a list of available flags, run "gdm -h"

      Supported device types:
      Primary:
      {}

      Auxiliary:
      {}

      Virtual:
      {}

      To explore available device functionality through the dynamic CLI, you will need a device
      attached to your host.

      Use "gdm man" to access static documentation (no devices necessary, but has limitations):
          "gdm man <device_type>" to see all functionality supported by a device.
          "gdm man <device_type> --deprecated" to see just deprecated functionality for a device.
          "gdm man <device_type> <class_attribute>" for device attribute documentation.
          "gdm man <device_type> <capability> <capability_attribute>" for capability attribute documentation.
    """).format(
        cls._indent_doc_lines(cls.get_supported_primary_device_types()),
        cls._indent_doc_lines(cls.get_supported_auxiliary_device_types()),
        cls._indent_doc_lines(cls.get_supported_virtual_device_types()))[1:-1]
    logger.info(description)

  def issue(self, identifier, **kwargs):
    """Alias for create_device with health checks enabled by default.

    Usage from the CLI: `gdm issue device-1234 - shell "echo 'foo'"`.

    Args:
      identifier (str): The identifier string (name) which specifies the
        device.
      **kwargs (dict): keyword-only arguments passed on to create_device:
        "log_file_name" (str) -- a string log file name to use for log
        results. "log_directory" (str) -- a directory path to use for storing
        log file. "make_device_ready" (str) -- health check setting ("on",
        "check_only", "off", "flash_build").

    Returns:
      Object: The device found and created by the identifier specified.
    """
    return self.create_device(
        identifier,
        log_file_name=kwargs.get("log_file_name"),
        log_directory=kwargs.get("log_directory"),
        make_device_ready=kwargs.get("make_device_ready", "on"))

  def issue_devices(self,
                    devices: Sequence[str],
                    attribute_name: str,
                    *method_args: Any,
                    timeout: float = parallel_utils.TIMEOUT_PROCESS,
                    **method_kwargs: Any) -> List[Any]:
    """Executes a device method or property in parallel for multiple devices.

    For example: 'gdm issue-devices device-1234,device-3456 reboot'.

    Args:
      devices: Device identifiers.
      attribute_name: Name of device method or property to execute in parallel.
        Can be nested. For example: "shell", "wifi.ssid".
      *method_args: Positional arguments to the device method.
      timeout: Maximum amount of seconds to allow parallel methods to complete.
      **method_kwargs: Keyword arguments to the device method.

    Returns:
      Results from parallel calls.
    """
    if isinstance(devices, str):
      devices = devices.split(",")
    return self._issue_devices(
        devices, attribute_name, timeout, method_args, method_kwargs)

  def issue_devices_all(self,
                        attribute_name: str,
                        *method_args: Any,
                        timeout: float = parallel_utils.TIMEOUT_PROCESS,
                        **method_kwargs: Any) -> List[Any]:
    """Executes a device method or property in parallel for connected devices.

    For example: 'gdm issue-devices-all reboot --no_wait=True'.

    Args:
      attribute_name: Name of device method or property to execute in parallel.
        Can be nested. For example: "shell", "wifi.ssid".
      *method_args: Positional arguments to the device method.
      timeout: Maximum amount of seconds to allow parallel methods to complete.
      **method_kwargs: Keyword arguments to the device method.

    Returns:
      Results from parallel calls.

    Raises:
      DeviceError: if no devices are connected.
    """
    devices = self.get_connected_devices()
    if not devices:
      raise errors.DeviceError("No devices are connected.")
    return self._issue_devices(
        devices, attribute_name, timeout, method_args, method_kwargs)

  def issue_devices_match(self,
                          match: str,
                          attribute_name: str,
                          *method_args: Any,
                          timeout: float = parallel_utils.TIMEOUT_PROCESS,
                          **method_kwargs: Any) -> List[Any]:
    """Executes a device method or property in parallel for matching devices.

    For example: 'gdm issue-devices-match raspberrypi* reboot --no_wait=True'.

    Args:
      match: Wildcard-supported string to match against device names, i.e.
        "raspberrypi*" will call provided method on all connected Raspberry Pis.
      attribute_name: Name of device method or property to execute in parallel.
        Can be nested. For example: "shell", "wifi.ssid".
      *method_args: Positional arguments to the device method.
      timeout: Maximum amount of seconds to allow parallel methods to complete.
      **method_kwargs: Keyword arguments to the device method.

    Returns:
      Results from parallel calls.

    Raises:
      DeviceError: if provided wildcard does not match any connected devices.
    """
    devices = fnmatch.filter(self.get_connected_devices(), match)
    if not devices:
      raise errors.DeviceError(f"No devices match {match!r}.")
    return self._issue_devices(
        devices, attribute_name, timeout, method_args, method_kwargs)

  def log(self, device_name, log_file_name=None, duration=2000):
    """Streams device logs to stdout.

    Args:
      device_name (str): device identifier.
      log_file_name (str): log_file_name. Used for testing purposes.
      duration (float): how long to stream logs for.

    Raises:
      DeviceError: if unable to initiate log file to stream in 10 seconds.
    """
    logger.info("Streaming logs for max {}s".format(duration))
    device = self.create_device(device_name, log_file_name=log_file_name)
    # Disable log rotation feature
    if hasattr(type(device), "switchboard"):
      device.switchboard.set_max_log_size(0)

    try:
      # Wait up to 10 seconds for log file to be created
      end_time = time.time() + MAX_TIME_TO_WAIT_FOR_INITATION
      while time.time() < end_time:
        if os.path.exists(device.log_file_name):
          break
        time.sleep(0.001)
      else:
        raise errors.DeviceError(
            "Streaming logs for {} failed. "
            "Log file not created within {} seconds".format(
                device_name, MAX_TIME_TO_WAIT_FOR_INITATION))

      start_time = time.time()
      end_time = start_time + duration
      # Open log file and process log file
      with codecs.open(
          device.log_file_name, "r", encoding="utf-8",
          errors="replace") as log_file:

        while time.time() < end_time:
          line = log_file.readline()
          if line:
            sys.stdout.write(line)
            sys.stdout.flush()
          else:
            time.sleep(0.001)

    finally:
      sys.stdout.flush()
      device.close()

  @classmethod
  def man(cls,
          device_type: Optional[str] = None,
          class_attr_name: Optional[str] = None,
          capability_attr_name: Optional[str] = None,
          deprecated: bool = False) -> None:
    """Prints documentation without reliance on device instance creation.

    Dynamic device documentation ("gdm <command> --help") is more complete, but
    requires a device attached to the host to generate documentation.

    Args:
      device_type: Device type.
      class_attr_name: Name of the class attribute to display documentation for.
      capability_attr_name: Name of the capability attribute to display
        documentation for.
      deprecated: Display only deprecated methods and properties.

    Raises:
      AttributeError: Requested attribute does not exist.
      TypeError: Requested capability attribute documentation for a
        non-capability.
    """
    # Handle both space-separated and dot-separated inputs:
    # gdm man cambrionix switch_power power_on
    # gdm man cambrionix.switch_power.power_on
    args = " ".join(
        arg for arg in (device_type, class_attr_name, capability_attr_name)
        if arg)
    args = args.replace(".", " ")
    # Allow both dashes and underscores for names.
    args = args.replace("-", "_")
    args_list = args.split()
    device_type, class_attr_name, capability_attr_name = (
        args_list + [None] * (3 - len(args_list)))

    if not device_type:
      cls.helpfull()
    else:
      device_class = cls.get_supported_device_class(device_type)
      if not class_attr_name:
        cls._man_device(device_type, device_class, deprecated)
      else:
        if not hasattr(device_class, class_attr_name):
          raise AttributeError(f"{device_type} ({device_class}) does not have "
                               f"attribute {class_attr_name!r}")
        class_attribute = getattr(device_class, class_attr_name)
        classification = cls._classify_attribute(class_attribute)
        class_attr_description = classification.value

        if classification == AttributeClassification.CAPABILITY:
          capability_classes = list(class_attribute.capability_classes)
          class_attribute = capability_classes[0]
          class_attr_description += f"; flavor: {class_attribute.__name__}"
          if len(capability_classes) > 1:
            _log_man_warning_for_multiple_flavors(
                capability_classes, class_attr_name, device_type,
                class_attribute)

        if not capability_attr_name:
          cls._man_class_attribute(
              device_class, class_attr_name, class_attr_description,
              class_attribute)
        else:
          if classification != AttributeClassification.CAPABILITY:
            raise TypeError(
                f"{device_type}.{class_attr_name} ({class_attribute}) is not a "
                "capability.")
          if not hasattr(class_attribute, capability_attr_name):
            raise AttributeError(
                f"{device_type}.{class_attr_name} ({class_attribute}) "
                f"does not have attribute {capability_attr_name!r}")
          capability_attribute = getattr(class_attribute, capability_attr_name)
          capability_attr_description = cls._classify_attribute(
              capability_attribute).value + f" of {class_attribute.__name__}"
          cls._man_capability_attribute(
              device_class, class_attr_name, capability_attr_name,
              capability_attr_description, capability_attribute)

  def print_usb_info(self):
    """Prints the usb_info dictionary in a human readable form."""
    values = usb_utils.get_address_to_usb_info_dict().values()
    logger.info("{} USB connections found.".format(len(values)))
    for num, usb_info_dict in enumerate(values):
      keys = usb_info_dict.get_properties()
      logger.info("Connection {}:".format(num))
      for key in keys:
        logger.info("\t{:15} {:15}".format(key,
                                           str(getattr(usb_info_dict, key))))

  def register(self, package_name: str) -> None:
    """Registers the given package with GDM CLI.

    Args:
      package_name: Name of the package to register. For example,
        "foo_extension_package" or "my_package.bar_devices".

    Note that this only registers the package for CLI usage. Tests and Python
    interpreter users must use package_registrar.register() instead.
    """
    registered_cli_packages = self.config.get("cli_extension_packages", [])
    if package_name not in registered_cli_packages:
      if package_registrar.import_and_register(package_name,
                                               include_cli_instructions=True):
        self._set_config_prop("cli_extension_packages",
                              registered_cli_packages + [package_name])
        logger.info(f"Registered package {package_name!r} with GDM CLI.")
    else:
      logger.info(
          f"Package {package_name!r} is already registered with GDM CLI.")

  def unregister(self, package_name: str) -> None:
    """Removes the given package from GDM CLI.

    Args:
      package_name: Name of the package to unregister. For example,
        "foo_extension_package" or "my_package.bar_devices".

    Note that this only removes the package from the CLI.
    """
    registered_cli_packages = self.config.get("cli_extension_packages", [])
    if package_name in registered_cli_packages:
      updated_packages = registered_cli_packages.copy()
      updated_packages.remove(package_name)
      self._set_config_prop("cli_extension_packages", updated_packages)
      logger.info(f"Removed package {package_name!r} from GDM CLI.")
    else:
      logger.info(f"Package {package_name!r} is not registered with GDM CLI.")

  def update_gdm(self):
    """Update GDM in this virtual environment.

    To update GDM in this virtual environment, call the GDM launcher script
    by using the full path as shown above.

    If no version is specified then GDM will be updated to the latest
    version available otherwise the version specified will be installed instead.
    """
    logger.info(textwrap.dedent(f"""
      Unable to update Gazoo Device Manager using this tool.
      If you want to update GDM call the GDM launcher script directly like this:
      {config.LAUNCHER_PATH} update-gdm [version]

      If after doing the above you see this message again, then you probably did a
      'sudo pip install gazoo-device' and overwrote the GDM launcher script. Please
      reinstall GDM.
    """))

  @classmethod
  def _classify_attribute(cls, class_attr):
    """Classifies the class attribute."""
    class_attr = decorators.unwrap(class_attr)
    if isinstance(class_attr, decorators.CapabilityProperty):
      return AttributeClassification.CAPABILITY
    if (hasattr(class_attr, "__deprecated__") or
        hasattr(getattr(class_attr, "fget", None), "__deprecated__")):
      if inspect.isroutine(class_attr):
        return AttributeClassification.DEPRECATED_METHOD
      if isinstance(class_attr, property):
        return AttributeClassification.DEPRECATED_PROPERTY
    if isinstance(class_attr, config.CLASS_PROPERTY_TYPES):
      return AttributeClassification.CLASS_PROPERTY
    if isinstance(class_attr, property):
      return AttributeClassification.PROPERTY
    if inspect.isroutine(class_attr):
      if hasattr(class_attr, "__health_check__"):
        return AttributeClassification.HEALTH_CHECK
      if class_attr.__name__.startswith("_"):
        return AttributeClassification.PRIVATE_METHOD
      return AttributeClassification.PUBLIC_METHOD
    return AttributeClassification.OTHER

  @classmethod
  def _indent_doc_lines(cls, doc_lines, indent=_DOC_INDENT_SIZE):
    """Indents docstring lines."""
    indent_str = " " * indent
    return "\n".join(indent_str + line for line in doc_lines)

  def _issue_devices(self,
                     devices: Sequence[str],
                     attribute_name: str,
                     timeout: float,
                     method_args: Tuple[Any, ...],
                     method_kwargs: Mapping[str, Any]) -> List[Any]:
    """Executes a device method in parallel on multiple devices.

    Args:
      devices: Device identifiers to execute the method on.
      attribute_name: Name of device method or property to execute in parallel.
      timeout: Maximum amount of seconds to allow parallel methods to complete.
      method_args: Positional arguments to the method.
      method_kwargs: Keyword arguments to the method.

    Returns:
      Results from the parallel method calls.
    """
    call_specs = [
        parallel_utils.CallSpec(_run_device_action, device, attribute_name,
                                method_args, method_kwargs)
        for device in devices]
    results, _ = parallel_utils.execute_concurrently(
        call_specs=call_specs, timeout=timeout, raise_on_process_error=True)
    return results

  @classmethod
  def _man_capability_attribute(cls,
                                device_class: Type[Any],
                                capability_name: str,
                                capability_attr_name: str,
                                description: str,
                                attribute: Any) -> None:
    """Prints capability attribute documentation.

    Invoked by 'man' with 3 args (device, capability, attribute).

    Args:
      device_class: Device class which contains the capability.
      capability_name: Name of the capability containing the attribute.
      capability_attr_name: Capability attribute to print documentation for.
      description: Description of the capability attribute.
      attribute: The capability attribute to display documentation for.
    """
    doc_title = (
        f"Manual for "
        f"{device_class.__name__}.{capability_name}.{capability_attr_name} "
        f"({description})\n")
    pydoc_lines = pydoc.render_doc(attribute).splitlines()
    doc = doc_title + "\n".join(pydoc_lines[1:])  # Replace pydoc's title
    logger.info(doc)

  @classmethod
  def _man_class_attribute(cls,
                           device_class: Type[Any],
                           class_attr_name: str,
                           description: str,
                           attribute: Any) -> None:
    """Prints class attribute documentation.

    Invoked by 'man' with 2 args (device, attribute).

    Args:
      device_class: Device class which contains the attribute.
      class_attr_name: Name of the class attribute to display documentation for.
      description: Description of the class attribute.
      attribute: The attribute to display documentation for.
    """
    doc_title = (f"Manual for {device_class.__name__}.{class_attr_name} "
                 f"({description})\n")
    pydoc_lines = pydoc.render_doc(attribute).splitlines()
    doc = doc_title + "\n".join(pydoc_lines[1:])  # Replace pydoc's title
    logger.info(doc)

  @classmethod
  def _man_device(cls,
                  device_type: str,
                  device_class: Type[Any],
                  deprecated: bool) -> None:
    """Prints supported device features.

    Args:
      device_type: Type of device.
      device_class: Device class to show documentation for.
      deprecated: Whether to display manuals of deprecated attributes.

    Invoked by 'man' with 1 argument (device type).
    """
    # Group device class attributes into properties, health checks, methods,
    # and capabilities
    capability_name_to_classes = {}
    property_names = []
    class_properties = []
    public_methods_names = []
    health_check_names = []
    deprecated_methods = []
    deprecated_properties = []
    for name, attribute in inspect.getmembers(device_class):
      classification = cls._classify_attribute(attribute)
      if classification == AttributeClassification.PROPERTY:
        property_names.append(name)
      elif classification == AttributeClassification.CLASS_PROPERTY:
        if not name.startswith("_"):
          class_properties.append(f"{name}: {attribute}")
      elif classification == AttributeClassification.HEALTH_CHECK:
        health_check_names.append(name)
      elif classification == AttributeClassification.PUBLIC_METHOD:
        public_methods_names.append(name)
      elif classification == AttributeClassification.CAPABILITY:
        capability_name_to_classes[name] = list(attribute.capability_classes)
      elif classification == AttributeClassification.DEPRECATED_METHOD:
        deprecated_methods.append(f"{name} ({attribute.__deprecated__})")
      elif classification == AttributeClassification.DEPRECATED_PROPERTY:

        # Parse alias from deprecated property docstring
        match = re.search(r'See "(?P<alias>.*)".', attribute.__doc__)
        if match:
          alias = match.group("alias")
          deprecated_properties.append(f"{name} ({alias})")

    # Generate a summary of supported capability methods and properties for each
    # capability
    capability_lines = []
    for cap_name, cap_classes in capability_name_to_classes.items():
      # There may be several flavors of a capability for a device class.
      for cap_class in cap_classes:
        capability_lines.append("{} ({})".format(cap_name, cap_class.__name__))
        methods_and_props = [
            name for name, member in inspect.getmembers(cap_class)
            if cls._classify_attribute(member) in _VISIBLE_CAPABILITY_ATTRIBUTES
        ]
        indented_lines = [
            " " * _DOC_INDENT_SIZE + line for line in methods_and_props
        ]
        capability_lines.extend(indented_lines)

    docs_capabilities = cls._indent_doc_lines(capability_lines)
    docs_methods = cls._indent_doc_lines(public_methods_names)
    docs_health_checks = cls._indent_doc_lines(health_check_names)
    docs_properties = cls._indent_doc_lines(property_names)
    docs_class_properties = cls._indent_doc_lines(class_properties)
    docs_deprecated_methods = cls._indent_doc_lines(deprecated_methods)
    docs_deprecated_properties = cls._indent_doc_lines(deprecated_properties)

    if deprecated:
      template = textwrap.dedent("""
        Deprecated manual for device type '{device_type}' (class {device_class})

        Deprecated methods:
        {deprecated_methods}

        Deprecated properties:
        {deprecated_properties}

        Use "gdm man {device_type} <deprecated_attribute>" for device attribute documentation.
      """)[1:-1]

    else:
      template = textwrap.dedent("""
        Manual for device type '{device_type}' (class {device_class})

        Owned (maintained) by {device_class._OWNER_EMAIL}

        Supported capabilities:
        {capabilities}

        Supported methods:
        {methods}

        Supported health check methods:
        {health_checks}

        Supported properties:
        {properties}

        Class properties:
        {class_properties}

        Use:
          "gdm man <device_type> <class_attribute>" for device attribute documentation.
          "gdm man <device_type> <capability> <capability_attribute>" for capability attribute documentation.
      """)[1:-1]

    doc = template.format(
        device_type=device_type,
        device_class=device_class,
        capabilities=docs_capabilities,
        methods=docs_methods,
        health_checks=docs_health_checks,
        properties=docs_properties,
        class_properties=docs_class_properties,
        deprecated_methods=docs_deprecated_methods,
        deprecated_properties=docs_deprecated_properties)
    logger.info(doc)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Methods for mobly controller.

Imported into init.

To use:
import gazoo_device
from mobly import base_test

class Test(base_test.TestCase):
  def set_up_class(self):
    self.devices = self.register_controller(gazoo_device)


  def setup_test(self):
    for device in self.devices:
      device.start_new_log(log_name_prefix=self.current_test_info.name)

  def teardown_test(self):
    for device in self.devices:
      device.make_device_ready()
"""
import logging
import os
from typing import Any, Dict, List, Optional, Sequence

from gazoo_device import custom_types
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device import manager

# Device properties for gazoo device analytics.
_DEVICE_PROPS = ("name", "device_type", "model", "platform", "serial_number",
                 "wifi_mac_address", "firmware_version", "firmware_branch",
                 "firmware_type", "alias", "communication_address",
                 "secondary_communication_address", "build_date",
                 "initial_code_name")
_LOGGER = gdm_logger.get_logger()
_MANAGER_INSTANCE = None


def _get_log_directory() -> Optional[str]:
  """Returns the path to the directory where logs should be stored.

  Returns:
    Directory path, or None if the default log directory should be used.
  """
  return getattr(
      logging, "log_path",  # Set by Mobly in base_test.BaseTestClass.run.
      os.getenv("TEST_UNDECLARED_OUTPUTS_DIR"))


def get_manager() -> manager.Manager:
  """Set up, as needed, and returns a manager instance."""
  global _MANAGER_INSTANCE
  if _MANAGER_INSTANCE is None:
    log_directory = _get_log_directory()
    if log_directory is not None:
      log_file = os.path.join(log_directory, "gdm.txt")
    else:
      log_file = None
    _MANAGER_INSTANCE = manager.Manager(
        log_directory=log_directory,
        gdm_log_file=log_file,
        stdout_logging=False)
  return _MANAGER_INSTANCE


def _set_auxiliary_props(properties: Dict[str, Any], device_name: str):
  """Sets props on device instance to the values in the config.

  Label is translated to Alias and dimensions are left out.

  Args:
    properties: Dictionary of device properties and their values.
    device_name: Id of the device instance.
  """
  for prop_name, value in properties.items():
    if prop_name in ["id", "dimensions"]:
      continue
    elif prop_name == "label":
      get_manager().set_prop(device_name, "alias", value)
    else:
      try:
        get_manager().set_prop(device_name, prop_name, value)
      except errors.DeviceError:
        _LOGGER.warning("%s unsettable for %s", prop_name, device_name)


def create(configs: List[Dict[str, Any]]) -> List[custom_types.Device]:
  """Creates gazoo device instances and returns them."""
  log_directory = _get_log_directory()
  devices = []
  for entry in configs:
    name = entry["id"]
    _set_auxiliary_props(entry, name)
    if entry.get("bypass_gdm_check") == "true":
      _LOGGER.info(f"bypass_gdm_check is set for {name}. "
                   "Skipping health checks")
      device = get_manager().create_device(
          name, log_directory=log_directory, make_device_ready="off")
    else:
      device = get_manager().create_device(name, log_directory=log_directory)
    devices.append(device)
  return devices


def get_info(devices: Sequence[custom_types.Device]) -> List[Dict[str, Any]]:
  """Returns persistent info and firmware version for each device."""
  info = []
  for device in devices:
    props = {}
    for device_prop in _DEVICE_PROPS:
      try:
        props[device_prop] = device.get_property(device_prop, raise_error=True)
      except AttributeError:
        props[device_prop] = "Undefined"
      except errors.DeviceError as e:
        props[device_prop] = repr(e)
    info.append(props)
  return info


def destroy(devices: List[custom_types.Device]) -> None:
  """Closes all created devices and manager."""
  for device in devices:
    device.close()
  global _MANAGER_INSTANCE
  if _MANAGER_INSTANCE is not None:
    _MANAGER_INSTANCE.close()
  _MANAGER_INSTANCE = None



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Enables logging from multiple processes.

Step 1: Initialize a queue in the main process.
Step 2: Pass that queue to a LoggingThread object.
Step 3: Add handlers (StreamHandler, FileHandler, etc) to the LoggingThread.
Step 4: In each process, initialize a QueueHandler and add it to a Logger.
Step 5: Send messages to the Logger.

Messages travel from the Logger, to the QueueHandler, into the LoggingThread's
child thread via the queue, then into the destination handlers.

Adapted from the implementation in Python 3.5.
"""
import gc
import logging
import sys
import threading
from typing import List, Optional

SYNC_TIMEOUT = 0.25
TERMINATE_TIMEOUT = 2


class _Sentinel(object):
  """Sentinels to put in the queue to signal certain events."""
  TERMINATE = 0
  SYNC = 1


class DisablePeriodicGC(object):
  """Context manager for disabling & re-enabling periodic garbage collection.

  Note: will not disable periodic GC if called during periodic garbage
  collection.
  See b/143321081 for why logging during __del__ can deadlock if periodic GC
  is not disabled.
  """
  is_periodic_gc_active = False

  @classmethod
  def _track_gc_state(cls, phase, _):
    if phase == "start":
      cls.is_periodic_gc_active = True
    elif phase == "stop":
      cls.is_periodic_gc_active = False

  def __enter__(self):
    """Enter the context manager & disable periodic GC."""
    self._was_periodic_gc_enabled = gc.isenabled()
    if self._was_periodic_gc_enabled and not self.is_periodic_gc_active:
      gc.disable()

  def __exit__(self, exc_type, exc_value, exc_traceback):
    """Exit the context manager & re-enable periodic GC (if it had been enabled before)."""
    if self._was_periodic_gc_enabled and not self.is_periodic_gc_active:
      gc.enable()


if hasattr(gc, "callbacks"):  # Available in Python 3 only
  gc.callbacks.append(DisablePeriodicGC._track_gc_state)


class QueueHandler(logging.Handler):
  """Receives all messages destined for the handlers in LoggingThread."""

  def __init__(self, queue):
    logging.Handler.__init__(self)
    self._queue = queue

  def emit(self, record):
    try:
      self.format(record)
      record.msg = record.message
      record.args = None
      record.exc_info = None

      with DisablePeriodicGC():
        self._queue.put_nowait(record)

    except Exception:
      self.handleError(record)


class LoggingThread(object):
  """Runs in main process and pulls log messages from the shared queue."""

  def __init__(self, queue):
    self._handlers = []
    self._propagate = False
    self._parent = None
    self._queue = queue
    self._thread = None
    self._synchonize_event = threading.Event()

  @property
  def handlers(self) -> List[logging.Handler]:
    """Active logging handlers."""
    return self._handlers

  def add_handler(self, handler):
    """Adds a logging handler to the LoggingThread.

    Args:
        handler (logging.Handler): logging handler to add to the logging
          queue handlers.  The handler will receive messages that are placed
          on the shared queue by QueueHandlers.
    """
    self._handlers.append(handler)

  def remove_handler(self, handler):
    """Removes the given logging handler from the LoggingThread."""

    try:
      self._handlers.remove(handler)
    except ValueError:
      pass

  @property
  def propagate(self) -> bool:
    """Whether the LoggingThread propagates messages to ancestor loggers."""
    return self._propagate

  @property
  def parent(self) -> Optional[logging.Logger]:
    """The LoggingThread's parent Logger, or None if not configured."""
    return self._parent

  def configure_propagate(
      self, propagate: bool, parent: logging.Logger) -> None:
    """Configures the LoggingThread's propagation setup."""

    self._propagate = propagate
    self._parent = parent

  def start(self):
    """Starts the child thread, which pulls messages from the queue."""

    self._thread = threading.Thread(
        target=self._run, args=[self._queue, self._synchonize_event])
    self._thread.daemon = True
    self._thread.start()

  def stop(self):
    """Signals to the child thread to stop by putting a sentinel on the queue.

    Raises:
      RuntimeError: Thread alive after attempt to stop it.
    """
    if self._thread:
      with DisablePeriodicGC():
        self._queue.put_nowait(_Sentinel.TERMINATE)
      self._thread.join(timeout=TERMINATE_TIMEOUT)

      if self._thread.is_alive():
        raise RuntimeError(
            "Failed to stop LoggingThread in {}s".format(TERMINATE_TIMEOUT))

    self._thread = None

  def sync(self, timeout=SYNC_TIMEOUT):
    """Put a sentinel in the message queue and block until the logging thread reaches it.

    Args:
        timeout (float): maximum time to wait for logging thread to reach
          the synchronization sentinel.

    Note: Does nothing in child (forked) processes because logging thread is
      not copied over by os.fork(). Also does nothing if called in any
      thread other than the main thread.
    """
    if (self._thread is not None and self._thread.is_alive() and
        threading.current_thread() == threading.main_thread()):
      with DisablePeriodicGC():
        self._queue.put_nowait(_Sentinel.SYNC)
      if not self._synchonize_event.wait(timeout):
        print(
            "Warning: Logging thread did not reach the synchronization sentinel in {}s"
            .format(timeout),
            file=sys.stderr)
      self._synchonize_event.clear()

  def _run(self, queue, synchronize_event):
    """Runs as a child thread.

    Args:
        queue (multiprocessing.Queue): multiprocess queue to receive log
          messages from.
        synchronize_event (threading.Event): event used for synchronization
          with the main execution thread.  Continuously receiving messages
          placed on the queue by QueueHandlers. Passes messages to each of
          the LoggingThread's handlers. Stops when _Sentinel.TERMINATE is found
          on the queue.
    """
    while True:
      record = queue.get()

      if record == _Sentinel.TERMINATE:
        break
      elif record == _Sentinel.SYNC:
        synchronize_event.set()
      else:
        # This is basically logging.Logger.callHandlers(), but handled from
        # within a LoggingThread. See b/227232743.
        curr_logger = self
        while curr_logger is not None:
          for handler in curr_logger.handlers:
            if record.levelno >= handler.level:
              handler.handle(record)
          curr_logger = curr_logger.parent if curr_logger.propagate else None



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Interface for a Matter OnOff light endpoint."""
import abc
from gazoo_device.capabilities.matter_endpoints.interfaces import endpoint_base


class OnOffLightBase(endpoint_base.EndpointBase, metaclass=abc.ABCMeta):
  """Matter On/Off light endpoint interface."""

  DEVICE_TYPE_ID = 0x0100

  @property
  @abc.abstractmethod
  def on_off(self):
    """Required cluster: ZCL on_off cluster."""

  @property
  @abc.abstractmethod
  def level(self):
    """Optional cluster: ZCL level cluster."""

  # TODO(b/209362086) Add the below clusters
  # @abc.abstractproperty
  # def scenes(self) -> None:
  #   """Required cluster: ZCL scenes cluster."""

  # @abc.abstractproperty
  # def groups(self) -> None:
  #   """Required cluster: ZCL groups cluster."""

  # @abc.abstractproperty
  # def occupancy(self) -> None:
  #   """Optional cluster: ZCL occupancy cluster."""



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Pigweed RPC implementation of the Matter Color Control cluster capability.
"""
from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.capabilities import matter_enums
from gazoo_device.capabilities.matter_clusters.interfaces import color_control_base
from gazoo_device.protos import attributes_service_pb2

logger = gdm_logger.get_logger()
ColorControlCluster = matter_enums.ColorControlCluster
INT8U_ATTRIBUTE_TYPE = attributes_service_pb2.AttributeType.ZCL_INT8U_ATTRIBUTE_TYPE


class ColorControlClusterPwRpc(color_control_base.ColorControlClusterBase):
  """Matter Color Control cluster capability."""

  @decorators.CapabilityLogDecorator(logger)
  def move_to_hue(self, hue: int, verify: bool = True) -> None:
    """The MoveToHue command.

    On receipt of this command, a device should move from its current hue
    to the given hue value.

    Args:
      hue: The hue that the device should move to.
      verify: If true, verifies the hue changes before returning.
    """
    previous_hue = self.current_hue

    self._write(
        endpoint_id=self._endpoint_id,
        cluster_id=ColorControlCluster.ID,
        attribute_id=ColorControlCluster.ATTRIBUTE_CURRENT_HUE,
        attribute_type=INT8U_ATTRIBUTE_TYPE,
        data_uint8=hue)

    if verify:
      if self.current_hue != hue:  # pylint: disable=comparison-with-callable
        raise errors.DeviceError(
            f"Device {self._device_name} current hue didn't change to {hue} "
            f"from {previous_hue}.")

  @decorators.CapabilityLogDecorator(logger)
  def move_to_saturation(self, saturation: int, verify: bool = True) -> None:
    """The MoveToSaturation command.

    On receipt of this command, a device should move from its current saturation
    to the given saturation value.

    Args:
      saturation: The saturation that the device should move to.
      verify: If true, verifies the hue changes before returning.
    """
    previous_saturation = self.current_saturation

    self._write(
        endpoint_id=self._endpoint_id,
        cluster_id=ColorControlCluster.ID,
        attribute_id=ColorControlCluster.ATTRIBUTE_CURRENT_SATURATION,
        attribute_type=INT8U_ATTRIBUTE_TYPE,
        data_uint8=saturation)

    if verify:
      if self.current_saturation != saturation:  # pylint: disable=comparison-with-callable
        raise errors.DeviceError(
            f"Device {self._device_name} current saturation didn't change to "
            f"{saturation} from {previous_saturation}.")

  @decorators.DynamicProperty
  def current_hue(self) -> int:
    """The CurrentHue attribute.

    The CurrentHue attribute contains the current hue value of the light.

    Returns:
      The current hue.
    """
    color_data = self._read(
        endpoint_id=self._endpoint_id,
        cluster_id=matter_enums.ColorControlCluster.ID,
        attribute_id=ColorControlCluster.ATTRIBUTE_CURRENT_HUE,
        attribute_type=INT8U_ATTRIBUTE_TYPE)
    return color_data.data_uint8

  @decorators.DynamicProperty
  def current_saturation(self) -> int:
    """The CurrentSaturation attribute.

    The CurrentSaturation attribute contains the current saturation value of
    the light.

    Returns:
      The current saturation.
    """
    color_data = self._read(
        endpoint_id=self._endpoint_id,
        cluster_id=matter_enums.ColorControlCluster.ID,
        attribute_id=ColorControlCluster.ATTRIBUTE_CURRENT_SATURATION,
        attribute_type=INT8U_ATTRIBUTE_TYPE)
    return color_data.data_uint8



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Function call retry utility."""
import time
from typing import Any, Callable, Mapping, Optional, Sequence, Type, TypeVar

from gazoo_device import errors


def _default_is_successful(_: Any) -> bool:
  return True


def not_func(val: Any) -> bool:
  """Returns True if bool(val) evaluates to False."""
  return not bool(val)


def is_true(val: Any) -> bool:
  """Returns True if bool(val) evaluates to True."""
  return bool(val)


_FuncReturnType = TypeVar("_FuncReturnType")


def retry(
    func: Callable[..., _FuncReturnType],
    func_args: Sequence[Any] = (),
    func_kwargs: Optional[Mapping[str, Any]] = None,
    is_successful: Callable[[_FuncReturnType], bool] = _default_is_successful,
    timeout: float = 10,
    interval: float = 1,
    reraise: bool = True,
    exc_type: Type[Exception] = errors.CommunicationTimeoutError
) -> _FuncReturnType:
  """Retries func() until it succeeds or timeout is reached.

  Success of execution of func() is determined by is_successful() function,
  which should return True on successful execution of func().

  Args:
    func: Function to execute.
    func_args: Positional arguments to the function.
    func_kwargs: Keyword arguments to the function.
    is_successful: Function which takes in the result of func() and returns
      whether function execution should be considered successful. To indicate
      success, return True. Defaults to always returning True.
    timeout: If no run of func() succeeds in this time period, raise an error.
    interval: How long to wait between attempts of func().
    reraise: Whether to re-raise exceptions in func() or not. If True, will
      re-raise any exceptions from func(). If False, considers execution of
      func() a failure if an Exception is raised. is_successful() will NOT be
      called if an Exception occurs.
    exc_type: Type of exception to raise when timeout is reached. Note that the
    class constructor will be called with just 1 argument.

  Returns:
    Return value of first successful func() call.

  Raises:
    Exception: if timeout is reached, or if an Exception occurs in func() with
    reraise=True.
  """
  if func_kwargs is None:
    func_kwargs = {}

  tried_times = 0
  start_time = time.time()
  end_time = start_time + timeout

  while time.time() < end_time:
    exception_occurred = False
    tried_times += 1
    try:
      func_result = func(*func_args, **func_kwargs)
    except Exception:  # pylint: disable=broad-except
      if reraise:
        raise
      else:
        exception_occurred = True

    if not exception_occurred and is_successful(func_result):
      return func_result

    time.sleep(interval)

  time_elapsed = time.time() - start_time
  raise exc_type("Timeout in {}s. Tried calling {} {} times.".format(
      time_elapsed, func.__name__, tried_times))



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Matter Temperature Sensor endpoint.

A Temperature Sensor device reports measurements of temperature.

This endpoint module corresponds to the
"Temperature Sensor" Matter device type which defines required and optional
clusters in the Matter spec 1.0.

A Temperature Sensor endpoint houses an instance of Temperature Sensor Matter
device type with the following defined clusters on this endpoint:

The required clusters for this endpoint: Temperature Measurement
"""
from gazoo_device import decorators
from gazoo_device.capabilities.matter_clusters import temperature_measurement_pw_rpc
from gazoo_device.capabilities.matter_endpoints.interfaces import temperature_sensor_base


class TemperatureSensorEndpoint(temperature_sensor_base.TemperatureSensorBase):
  """Matter Temperature Sensor endpoint."""

  @decorators.CapabilityDecorator(
      temperature_measurement_pw_rpc.TemperatureMeasurementClusterPwRpc)
  def temperature_measurement(
      self
  ) -> temperature_measurement_pw_rpc.TemperatureMeasurementClusterPwRpc:
    """Matter Occupancy cluster instance."""
    return self.cluster_lazy_init(
        temperature_measurement_pw_rpc.TemperatureMeasurementClusterPwRpc)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines a process transport by wrapping the subprocess.Popen() constructor."""
import select
import subprocess
import time
import six
from gazoo_device.switchboard import transport_properties
from gazoo_device.switchboard.transports import transport_base


class ProcessTransport(transport_base.TransportBase):
  """Perform communication using a subprocess with the command and args specified."""

  def __init__(self,
               command,
               args="",
               auto_reopen=False,
               open_on_start=True,
               working_directory=None):
    """Initialize the ProcessTransport object with the given process properties.

    Args:
        command (str): command to execute
        args (str): list of additional args to pass to command
        auto_reopen (bool): flag indicating transport should be reopened if
          unexpectedly closed.
        open_on_start (bool): flag indicating transport should be open on
          TransportProcess start.
        working_directory (string): Current working directory.
    """

    super(ProcessTransport, self).__init__(auto_reopen, open_on_start)
    self._args = [command] + args.split()

    self._properties.update({
        transport_properties.CLOSE_FDS: True,
    })
    self._process = None
    self._cwd = working_directory

  def is_open(self):
    """Returns True if transport is currently open.

    Returns:
        bool: True if transport is currently open, False otherwise.

    Note:
        If the process unexpected exits this method will return False.
    """
    return (hasattr(self, "_process") and self._process is not None and
            self._process.poll() is None)

  def _open(self):
    """Opens or reopens the process using the current property values."""
    if self._is_ready_to_open():
      self._process = subprocess.Popen(
          self._args,
          bufsize=0,
          close_fds=self._properties[transport_properties.CLOSE_FDS],
          stdin=subprocess.PIPE,
          stdout=subprocess.PIPE,
          stderr=subprocess.STDOUT,
          cwd=self._cwd)

  def _close(self):
    """The close() implementation checks if process is open differently than is_open().

    Hence this method is unused, but has to be overridden.
    """
    pass

  def close(self):
    """Closes the process."""
    if hasattr(self, "_process") and self._process:
      self._process.stdin.close()
      self._process.stdout.close()
      if self._process.poll() is None:
        self._process.terminate()
        while self._process.poll() is None:
          time.sleep(0.001)
      self._process = None

  def _read(self, size=1, timeout=None):
    """Returns bytes read up to max_bytes within timeout in seconds specified.

    Args:
        size (int): maximum number of bytes to read within timeout seconds
        timeout (float): maximum seconds to wait to read bytes or
          indefinitely
                         if timeout=None. Default: None

    Returns:
        str: bytes read from transport or None if no bytes were read
    """
    if timeout:
      return self._read_non_blocking(size, timeout)
    return self._read_data(size)

  def _write(self, data, timeout=None):
    """Writes data provided within timeout in seconds specified.

    Args:
        data (str): bytes to be written out within timeout seconds
        timeout (float): maximum seconds to wait to write bytes or
          indefinitely
                         if timeout=None. Default: None

    Returns:
        int: number of bytes written or None if no bytes were written
    """
    if isinstance(data, six.text_type):
      data = data.encode("utf-8", errors="replace")

    if timeout:
      return self._write_non_blocking(data, timeout)
    return self._write_data(data)

  def _is_ready_to_open(self):
    return True

  def _read_data(self, size):
    return self._process.stdout.read(size)

  def _read_non_blocking(self, size, timeout):
    count = 0
    result = b""
    end_time = time.time() + timeout
    while count < size and time.time() < end_time:
      time_left = end_time - time.time()
      readable, _, _ = select.select([self._process.stdout], [], [],
                                     0 if time_left < 0 else time_left)
      for stream in readable:
        if stream is self._process.stdout:
          count += 1
          result += self._read_data(1)
    return result

  def _write_non_blocking(self, data, timeout):
    count = 0
    size = len(data)
    end_time = time.time() + timeout
    while count < size and time.time() < end_time:
      time_left = end_time - time.time()
      _, writable, _ = select.select([], [self._process.stdin], [],
                                     0 if time_left < 0 else time_left)
      for stream in writable:
        if stream is self._process.stdin:
          count += self._write_data(data[count])
    return count

  def _write_data(self, data):
    if isinstance(data, six.integer_types):
      data = six.int2byte(data)
    self._process.stdin.write(data)
    self._process.stdin.flush()
    return len(data)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""adb flavor of file transfer capability."""
import os

from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.capabilities.interfaces import file_transfer_base
from gazoo_device.utility import adb_utils

logger = gdm_logger.get_logger()


class FileTransferAdb(file_transfer_base.FileTransferBase):
  """adb file transfer capability."""

  def __init__(self, device_name, add_log_note_fn, communication_address):
    """Initialize adb file transfer capability.

    Args:
        device_name (str): name of the device using the capability.
        add_log_note_fn (func): add output to log file.
        communication_address (str): device serial number
    """
    super(FileTransferAdb, self).__init__(device_name=device_name)
    self._add_log_note_fn = add_log_note_fn
    self._communication_address = communication_address

  @decorators.CapabilityLogDecorator(logger)
  def recv_file_from_device(self, src, dest="./"):
    """Copies srcs on device to dest on host.

    Args:
        src (str): file path on device to copy to host.
        dest (str): destination path on host computer.

    Raises:
        DeviceError: if destination directory doesn't exist or copy failed.

    Note:
        If no dest is provided, the file will be copied to the current
        working directory on the host computer.
    """
    destination_dir = os.path.dirname(dest)
    if destination_dir != "." and not os.path.exists(destination_dir):
      raise errors.DeviceError(
          "Device {} receive from device failed. "
          "Destination directory {} doesn't appear to exist.".format(
              self._device_name, destination_dir))

    logger.info("%s receiving from device. Source: %s Destination: %s",
                self._device_name, src, dest)
    self._check_adb_mode(self._communication_address)
    output = adb_utils.pull_from_device(
        self._communication_address, src, destination_path=dest)
    for line in output.splitlines():
      self._add_log_note_fn(line + "\n")
    if not os.path.exists(dest):
      raise errors.DeviceError("Final file {} does not exist".format(dest))

  @decorators.CapabilityLogDecorator(logger)
  def send_file_to_device(self, src, dest):
    """Copies src from host to dest on the device.

    Args:
        src (str): local file path on host computer.
        dest (str): file path on device where the file should be copied to.

    Raises:
        DeviceError: if source file doesn't exist or copy failed.
    """
    if not os.path.exists(src):
      raise errors.DeviceError("Device {} send to device failed. "
                               "Source file {} doesn't appear to exist.".format(
                                   self._device_name, src))

    logger.info("%s sending file(s) to device. Source: %s Destination: %s",
                self._device_name, src, dest)
    self._check_adb_mode(self._communication_address)
    try:
      output = adb_utils.push_to_device(self._communication_address, src, dest)
      for line in output.splitlines():
        self._add_log_note_fn(line + "\n")
    except RuntimeError as err:
      raise errors.DeviceError("Unable to copy {} to {} on device. "
                               "Error: {!r}".format(src, dest, err))

  def _check_adb_mode(self, adb_serial):
    """Checks that the adb serial of the device is available for use.

    Args:
        adb_serial (str): adb serial of the device.

    Raises:
        DeviceError: device is not available in adb device list.
    """
    if not adb_utils.is_adb_mode(adb_serial):
      raise errors.DeviceError(
          "{}'s adb serial {} is not in adb device list.".format(
              self._device_name, adb_serial))



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: descriptor_service.proto
# pylint: skip-file
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='descriptor_service.proto',
  package='chip.rpc',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x18\x64\x65scriptor_service.proto\x12\x08\x63hip.rpc\"!\n\nDeviceType\x12\x13\n\x0b\x64\x65vice_type\x18\x01 \x01(\r\"\x1d\n\x07\x43luster\x12\x12\n\ncluster_id\x18\x01 \x01(\r\"\x1c\n\x08\x45ndpoint\x12\x10\n\x08\x65ndpoint\x18\x01 \x01(\r2\xf7\x01\n\nDescriptor\x12>\n\x0e\x44\x65viceTypeList\x12\x12.chip.rpc.Endpoint\x1a\x14.chip.rpc.DeviceType\"\x00\x30\x01\x12\x37\n\nServerList\x12\x12.chip.rpc.Endpoint\x1a\x11.chip.rpc.Cluster\"\x00\x30\x01\x12\x37\n\nClientList\x12\x12.chip.rpc.Endpoint\x1a\x11.chip.rpc.Cluster\"\x00\x30\x01\x12\x37\n\tPartsList\x12\x12.chip.rpc.Endpoint\x1a\x12.chip.rpc.Endpoint\"\x00\x30\x01\x62\x06proto3'
)




_DEVICETYPE = _descriptor.Descriptor(
  name='DeviceType',
  full_name='chip.rpc.DeviceType',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='device_type', full_name='chip.rpc.DeviceType.device_type', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=38,
  serialized_end=71,
)


_CLUSTER = _descriptor.Descriptor(
  name='Cluster',
  full_name='chip.rpc.Cluster',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='cluster_id', full_name='chip.rpc.Cluster.cluster_id', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=73,
  serialized_end=102,
)


_ENDPOINT = _descriptor.Descriptor(
  name='Endpoint',
  full_name='chip.rpc.Endpoint',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='endpoint', full_name='chip.rpc.Endpoint.endpoint', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=104,
  serialized_end=132,
)

DESCRIPTOR.message_types_by_name['DeviceType'] = _DEVICETYPE
DESCRIPTOR.message_types_by_name['Cluster'] = _CLUSTER
DESCRIPTOR.message_types_by_name['Endpoint'] = _ENDPOINT
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

DeviceType = _reflection.GeneratedProtocolMessageType('DeviceType', (_message.Message,), {
  'DESCRIPTOR' : _DEVICETYPE,
  '__module__' : 'descriptor_service_pb2'
  # @@protoc_insertion_point(class_scope:chip.rpc.DeviceType)
  })
_sym_db.RegisterMessage(DeviceType)

Cluster = _reflection.GeneratedProtocolMessageType('Cluster', (_message.Message,), {
  'DESCRIPTOR' : _CLUSTER,
  '__module__' : 'descriptor_service_pb2'
  # @@protoc_insertion_point(class_scope:chip.rpc.Cluster)
  })
_sym_db.RegisterMessage(Cluster)

Endpoint = _reflection.GeneratedProtocolMessageType('Endpoint', (_message.Message,), {
  'DESCRIPTOR' : _ENDPOINT,
  '__module__' : 'descriptor_service_pb2'
  # @@protoc_insertion_point(class_scope:chip.rpc.Endpoint)
  })
_sym_db.RegisterMessage(Endpoint)



_DESCRIPTOR = _descriptor.ServiceDescriptor(
  name='Descriptor',
  full_name='chip.rpc.Descriptor',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=135,
  serialized_end=382,
  methods=[
  _descriptor.MethodDescriptor(
    name='DeviceTypeList',
    full_name='chip.rpc.Descriptor.DeviceTypeList',
    index=0,
    containing_service=None,
    input_type=_ENDPOINT,
    output_type=_DEVICETYPE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ServerList',
    full_name='chip.rpc.Descriptor.ServerList',
    index=1,
    containing_service=None,
    input_type=_ENDPOINT,
    output_type=_CLUSTER,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ClientList',
    full_name='chip.rpc.Descriptor.ClientList',
    index=2,
    containing_service=None,
    input_type=_ENDPOINT,
    output_type=_CLUSTER,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='PartsList',
    full_name='chip.rpc.Descriptor.PartsList',
    index=3,
    containing_service=None,
    input_type=_ENDPOINT,
    output_type=_ENDPOINT,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_DESCRIPTOR)

DESCRIPTOR.services_by_name['Descriptor'] = _DESCRIPTOR

# @@protoc_insertion_point(module_scope)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implementation of the switch_power_usb_default capability."""
from typing import Any, Callable, Dict

from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.capabilities.interfaces import switch_power_base

logger = gdm_logger.get_logger()

SYNC = "sync"
OFF = "off"


class SwitchPowerUsbDefault(switch_power_base.SwitchPowerBase):
  """Definition of the switch_power_usb_default capability."""
  _REQUIRED_COMMANDS = ["POWER_ON", "POWER_OFF", "GET_MODE"]
  _REQUIRED_REGEXS = ["GET_MODE_REGEX"]

  def __init__(self,
               shell_fn: Callable[..., Any],
               regex_shell_fn: Callable[..., Any],
               command_dict: Dict[str, str],
               regex_dict: Dict[str, str],
               device_name: str,
               serial_number: str,
               total_ports: int):
    """Initializes an instance of SwitchPowerUsbDefault capability.

    Args:
      shell_fn: shell() function of the device.
      regex_shell_fn: shell_with_regex() function of the device.
      command_dict: A dictionary containing the command used for each method.
      regex_dict: Response regexes to use to when identifying command responses.
      device_name: name of the device this capability is attached to.
      serial_number: serial number of device this capability is attached to.
      total_ports: Number of ports on the device.
    """
    super().__init__(device_name=device_name)
    self._shell_fn = shell_fn
    self._regex_shell_fn = regex_shell_fn
    self._command_dict = command_dict
    self._regex_dict = regex_dict
    self._serial_number = serial_number
    self._total_ports = total_ports

    self.validate_required_keys(self._REQUIRED_COMMANDS, command_dict,
                                "command_dict")
    self.validate_required_keys(self._REQUIRED_REGEXS, regex_dict, "regex_dict")

  @decorators.PersistentProperty
  def supported_modes(self):
    """Get the USB power modes supported by the USB hub."""
    return [OFF, SYNC]

  @decorators.PersistentProperty
  def total_ports(self):
    """Get the number of device ports.

    Returns:
        int: the number of device ports
    """
    return int(self._total_ports)

  def get_all_ports_mode(self):
    """Gets the USB mode for all ports on this hub.

    Returns:
      list: Returns a list of port modes with port number as index.
    """
    mode_list = []
    for port in range(1, self._total_ports + 1):
      mode_list.append(self.get_mode(port=port))
    return mode_list

  def get_mode(self, port):
    """Get the USB mode for specified port.

    Args:
      port (int): The port to get mode.

    Returns:
        str: USB port mode settings 'off', 'sync'

    Raises:
      DeviceError: invalid port.
    """
    port = int(port)
    self._validate_port("get_mode", port)
    cmd = self._command_dict["GET_MODE"].format(self._serial_number, port)
    result = self._regex_shell_fn(
        command=cmd,
        regex=self._regex_dict["GET_MODE_REGEX"],
        command_name="get_mode",
        tries=5)
    if "ON" in result.upper():
      return SYNC
    return OFF

  @decorators.CapabilityLogDecorator(logger)
  def power_on(self, port, data_sync=True):
    """This command powers on the port specified.

    Args:
        port (int): identifying which hub port to power on
        data_sync (bool): Ignored in default implementation

    Raises:
       DeviceError: if data_sync is False or if a valid port is not
       provided.
    """
    port = int(port)
    if not data_sync:
      raise errors.DeviceError(
          "Device {} {} power_on failed. "
          "data_sync=False not supported in default implementation.".format(
              self._device_name, self._serial_number))

    self._validate_port("power_on", port)
    logger.debug("{} Powering on usb port {}".format(self._device_name, port))
    self._shell_fn(self._command_dict["POWER_ON"].format(
        self._serial_number, port))

  @decorators.CapabilityLogDecorator(logger)
  def power_off(self, port):
    """This command powers off the port specified.

    Args:
        port (int): identifies which hub port to power off

    Raises:
       DeviceError: if operation failed or if a valid port is not provided.
    """
    port = int(port)
    self._validate_port("power_off", port)
    logger.debug("{} Powering off usb port {}".format(self._device_name, port))
    self._shell_fn(self._command_dict["POWER_OFF"].format(
        self._serial_number, port))

  @decorators.CapabilityLogDecorator(logger)
  def set_all_ports_mode(self, mode):
    """Sets all USB hub ports to the mode specified.

    Args:
      mode (str): USB hub mode to set. The mode must be in one of the
        supported_modes Example, 'off', 'sync'

    Raises:
      DeviceError: invalid mode.
    """
    self._validate_mode(mode)
    for port in range(1, self._total_ports + 1):
      self.set_mode(mode=mode, port=port)

  @decorators.CapabilityLogDecorator(logger)
  def set_mode(self, mode, port):
    """Sets the specified USB port to the mode specified.

    Args:
      mode (str): USB mode to set. Options: 'off', 'sync'
      port (int): The port to set.

    Raises:
      DeviceError: invalid mode or port.
    """
    port = int(port)
    mode = mode.lower()
    self._validate_port("set_mode", port)
    self._validate_mode(mode)

    if mode in OFF:
      self.power_off(port)
    else:  # "sync"
      self.power_on(port)

  def _validate_mode(self, mode):
    """Verify mode given resides in the valid mode list.

    Args:
        mode (str): mode e.g. 'off', 'sync'

    Raises:
        DeviceError: if mode given is not found in the valid mode list.
    """
    if mode not in list(self.supported_modes):
      raise errors.DeviceError(
          "Mode {} not found in supported modes {!r}".format(
              mode, self.supported_modes))

  def _validate_port(self, method_name, port):
    """Verify port given is within the bounds defined.

    Args:
        method_name (str): name of method requesting validation.
        port (int): value provided must be >= 1 and <= _total_ports.

    Raises:
        DeviceError: if port given is not within the bounds defined.
    """
    if not port or port > self._total_ports:
      raise errors.DeviceError("Device {} {} failed. "
                               "Port {} is invalid.".format(
                                   self._device_name, method_name, port))



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implementation of the embedded script capability for DLI powerswitch.

Allows more precision to operations invloving time senstivity to avoid
any latencies incured due to REST API calls over DLI powerswitch.
"""
from typing import Any, Callable, Sequence

from gazoo_device import decorators
from gazoo_device import gdm_logger
from gazoo_device.capabilities.interfaces import embedded_script_base

logger = gdm_logger.get_logger()

API_ENDPOINTS = {
    "GET_RUNNING_THREADS": "script/threads/",
    "START_SCRIPT": "script/start/",
    "TERMINATE_SCRIPT": "script/stop/"
}
HEADERS = {
    "Accept": "application/json",
    "Content-Type": "application/json",
    "X-CSRF": "x"
}


class EmbeddedScriptDliPowerswitch(embedded_script_base.EmbeddedScriptBase):
  """Implementation of the embedded_script capability."""

  def __init__(self, ip_address: str, base_url: str, device_name: str,
               http_fn: Callable[..., str]):
    """Initializes an instance of the embedded script capability.

    Args:
        ip_address: The IP address of the Powerswitch.
        base_url: Base url for REST api.
        device_name: Name of the device this capability is attached.
        http_fn: A function to send GET and POST http commands.
    """
    super().__init__(device_name=device_name)
    self._ip_address = ip_address
    self._base_url = base_url
    self._http_fn = http_fn

  def get_current_running_threads(self) -> str:
    """Returns current running threads over powerswitch.

    Returns:
      threads: Formatted with thread ids and script name. Response received
        from callable is expected to have JSON text. If there are threads
        running response like '{9:{label:test_script (admin via web ui) ON 2}}'
        can be expected. If no threads running an empty JSON text '{}' can be
        expected
    """

    threads_url = (self._base_url.format(ip=self._ip_address) +
                   API_ENDPOINTS["GET_RUNNING_THREADS"])
    threads = self._http_fn(method="GET", url=threads_url, headers=HEADERS)
    return threads

  @decorators.CapabilityLogDecorator(logger)
  def run(
      self, script_name: str, script_args: Sequence[Any] = ()) -> str:
    """Runs the embedded script over powerswitch.

    Script needs to be deployed over powerswitch manually before using run.
    Method returns numeric value for thread id in form of string.
    example - '21'. The same id can be passed to terminate method as it is
    to stop the execution of script .
    Args:
      script_name: Name of the script to run.
      script_args: Sequence of arguments if any consumed by script in order.

    Returns:
      str: Formatted GET/POST HTTP response.
    """
    script_url = (self._base_url.format(ip=self._ip_address) +
                  API_ENDPOINTS["START_SCRIPT"])
    script_arguments = ",".join(map(str, script_args))
    data = [{
        "user_function":
            script_name,
        "source":
            f"{script_name}({script_arguments})"
    }]
    response = self._http_fn(
        method="POST", url=script_url, headers=HEADERS, json_data=data)
    return response

  @decorators.CapabilityLogDecorator(logger)
  def terminate(self, thread_id: str = "all") -> str:
    """Terminates the thread executing the script.

    Terminates thread with provided thread_id, if no thread_id provided
    terminates all threads under execution.

    Args:
      thread_id: Id of thread executing the script.
        Defaults to 'all', terminating all threads under execution.

    Returns:
      str: Formatted GET/POST HTTP response.
    """
    script_url = (self._base_url.format(ip=self._ip_address) +
                  API_ENDPOINTS["TERMINATE_SCRIPT"])
    data = [thread_id]
    response = self._http_fn(
        method="POST", url=script_url, headers=HEADERS, json_data=data)
    return response




# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Matter controller capability using chip-tool command.

See
https://github.com/project-chip/connectedhomeip/tree/master/examples/chip-tool
for chip-tool usage and command references.
"""

import ast
import binascii
from typing import Any, Callable, Optional, Sequence

from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.capabilities.interfaces import matter_controller_base
import immutabledict

logger = gdm_logger.get_logger()

_CHIP_TOOL_BINARY_PATH = "/usr/local/bin/chip-tool"
_HEX_PREFIX = "hex:"

_COMMANDS = immutabledict.immutabledict({
    "READ_CLUSTER_ATTRIBUTE":
        "{chip_tool} {cluster} read {attribute} {node_id} {endpoint_id}",
    "WRITE_CLUSTER_ATTRIBUTE":
        "{chip_tool} {cluster} write {attribute} {value} {node_id} "
        "{endpoint_id}",
    "SEND_CLUSTER_COMMMAND":
        "{chip_tool} {cluster} {command} {arguments} {node_id} {endpoint_id}",
    "COMMISSION_OVER_BLE_WIFI":
        "{chip_tool} pairing ble-wifi {node_id} {ssid} {password} {setup_code} "
        "{long_discriminator}",
    "COMMISSION_OVER_BLE_THREAD":
        "{chip_tool} pairing ble-thread {node_id} {operational_dataset} "
        "{setup_code} {long_discriminator}",
    "COMMISSION_ON_NETWORK":
        "{chip_tool} pairing onnetwork {node_id} {setup_code}",
    "COMMISSION_ON_NETWORK_LONG":
        "{chip_tool} pairing onnetwork-long {node_id} {setup_code} "
        "{long_discriminator}",
    "DECOMMISSION":
        "{chip_tool} pairing unpair {node_id}",
    "CHIP_TOOL_VERSION":
        "cat ~/.matter_sdk_version",
    "WRITE_CHIP_TOOL_VERSION":
        "echo {chip_tool_version} > ~/.matter_sdk_version",
})

_REGEXES = immutabledict.immutabledict({
    "READ_CLUSTER_ATTRIBUTE_RESPONSE":
        r"CHIP:DMG:\s+Data = (\w+)",
    "WRITE_CLUSTER_ATTRIBUTE_RESPONSE":
        r"CHIP:DMG:\s+status = (0[xX][0-9a-fA-F]+)",
    "SEND_CLUSTER_COMMAND_RESPONSE":
        r"Received Command Response Status for Endpoint=\d+ "
        r"Cluster=0[xX][0-9a-fA-F_]+ Command=0[xX][0-9a-fA-F_]+ "
        r"Status=(0[xX][0-9a-fA-F]+)",
    "COMMAND_FAILURE":
        r"Run command failure: (.*)",
    "COMMISSION_SUCCESS":
        "(CHIP:TOO: Device commissioning completed with success)",
    "DECOMMISSION_COMPLETE":
        "(CHIP:DL: System Layer shutdown)",
})

_TIMEOUTS = immutabledict.immutabledict({
    "COMMISSION": 60,
    "SEND_CLUSTER_COMMAND": 30,
})


def _str_to_hex(value: str) -> str:
  """Formats a string into chip-tool compatible CLI argument.

  If the string starts with "hex:" prefix, then this function
  returns the value as is. Otherwise, the string is converted
  into its two-digit hex numbers form and prefixed with "hex:".
  For example, "password" is converted to "hex:70617373776f7264".

  See
  https://github.com/project-chip/connectedhomeip/tree/master/examples/chip-tool#using-the-client-to-commission-a-device
  for example usage.

  Args:
    value: The string to be converted.

  Returns:
    Value in its two-digit hex numbers form with "hex:" prefix.
  """
  if value.startswith(_HEX_PREFIX):
    return value

  hex_str = binascii.hexlify(str.encode(value))
  return f"{_HEX_PREFIX}{hex_str.decode()}"


class MatterControllerChipTool(matter_controller_base.MatterControllerBase):
  """Matter controller capability using chip-tool CLI commands."""

  def __init__(self,
               device_name: str,
               shell_fn: Callable[..., str],
               regex_shell_fn: Callable[..., str],
               send_file_to_device: Callable[[str, str], None],
               chip_tool_path: str = _CHIP_TOOL_BINARY_PATH):
    """Creates an instance of MatterControllerChipTool capability.

    Args:
      device_name: Name of the device this capability is attached to.
      shell_fn: Bound 'shell' method of the device class instance.
      regex_shell_fn: Bound 'shell_with_regex' method of the device class
        instance.
      send_file_to_device: Bound 'send_file_to_device' method of the device's
        file transfer capability instance.
      chip_tool_path: Path to chip-tool binary on the device.
    """
    super().__init__(device_name)

    self._chip_tool_path = chip_tool_path
    self._shell_with_regex = regex_shell_fn
    self._shell = shell_fn
    self._send_file_to_device = send_file_to_device

  @decorators.DynamicProperty
  def version(self) -> str:
    """Matter SDK version of the controller."""
    return self._shell(_COMMANDS["CHIP_TOOL_VERSION"])

  @decorators.CapabilityLogDecorator(logger)
  def commission(self,
                 node_id: int,
                 setup_code: str,
                 long_discriminator: Optional[int] = None,
                 ssid: Optional[str] = None,
                 password: Optional[str] = None,
                 operational_dataset: Optional[str] = None) -> None:
    """Commissions a device into the controller's fabric.

    Commissioning protocol is based on specified arguments:
      - When operational dataset is provided, pairs the device over ble-thread.
      - When Wi-Fi SSID and password are provided, pairs the device over
        ble-wifi.
      - Otherwise, discover the devices on the network and pairs with the first
        one that matches the setup code and long discriminator if one is
        specified.

    Args:
      node_id: Node ID to assign to the node being commissioned.
      setup_code: Set up PIN code of the remote device.
      long_discriminator: Discriminator of the remote device.
      ssid: Wi-Fi SSID either as a string, or in the form hex:XXXXXXXX where the
        bytes of the SSID are encoded as two-digit hex numbers.
      password: Wi-Fi password, either as a string or as hex data.
      operational_dataset: Thread dataset in base-64. This argument is mutually
        exclusive with ssid and password.
    """
    if ssid and not password:
      raise ValueError("Wi-Fi password is not specified.")

    if operational_dataset:
      command = _COMMANDS["COMMISSION_OVER_BLE_THREAD"]
      operational_dataset = _str_to_hex(operational_dataset)
    elif ssid:
      command = _COMMANDS["COMMISSION_OVER_BLE_WIFI"]
      ssid = _str_to_hex(ssid)
      password = _str_to_hex(password)
    elif long_discriminator is not None:
      # Commission the first device found on the network with provided
      # setup code and long discriminator.
      command = _COMMANDS["COMMISSION_ON_NETWORK_LONG"]
    else:
      # Commission the first device found on the network with provided
      # setup code.
      command = _COMMANDS["COMMISSION_ON_NETWORK"]

    command = command.format(
        chip_tool=self._chip_tool_path,
        node_id=node_id,
        setup_code=setup_code,
        long_discriminator=long_discriminator,
        ssid=ssid,
        password=password,
        operational_dataset=operational_dataset,
    )
    self._shell_with_regex(
        command,
        _REGEXES["COMMISSION_SUCCESS"],
        raise_error=True,
        timeout=_TIMEOUTS["COMMISSION"])

  @decorators.CapabilityLogDecorator(logger)
  def decommission(self, node_id: int) -> None:
    """Forgets a commissioned device with the given node id.

    Args:
      node_id: Assigned node id to decommission.
    """
    command = _COMMANDS["DECOMMISSION"].format(
        chip_tool=self._chip_tool_path, node_id=node_id)
    self._shell_with_regex(
        command, _REGEXES["DECOMMISSION_COMPLETE"], raise_error=True)

  def read(self, node_id: int, endpoint_id: int, cluster: str,
           attribute: str) -> Any:
    """Reads a cluster's attribute for the given node id and endpoint.

    Only primitive attribute values (integer, float, boolean and string)
    are supported.

    Args:
      node_id: Node ID assigned to the commissioned end device.
      endpoint_id: Endpoint ID within the node to read attribute from.
      cluster: Name of the cluster to read the attribute value from.
      attribute: Name of the cluster attribute to read.

    Returns:
      Attribute value of the cluster.
    """
    command = _COMMANDS["READ_CLUSTER_ATTRIBUTE"].format(
        chip_tool=self._chip_tool_path,
        node_id=node_id,
        endpoint_id=endpoint_id,
        cluster=cluster,
        attribute=attribute,
    )
    response = self._shell_with_regex(
        command, _REGEXES["READ_CLUSTER_ATTRIBUTE_RESPONSE"], raise_error=True)

    try:
      return ast.literal_eval(response)
    except ValueError:
      if response.lower() in ["true", "false"]:
        return response.lower() == "true"
    return response

  @decorators.CapabilityLogDecorator(logger)
  def write(self, node_id: int, endpoint_id: int, cluster: str, attribute: str,
            value: Any) -> None:
    """Writes a cluster's attribute for the given node id and endpoint.

    Args:
      node_id: Node ID assigned to the commissioned end device.
      endpoint_id: Endpoint ID within the node to write attribute to.
      cluster: Name of the cluster to write the attribute value to (e.g. onoff).
      attribute: Name of the cluster attribute to write (e.g. on-time).
      value: New attribute value to update the cluster with.
    """
    command = _COMMANDS["WRITE_CLUSTER_ATTRIBUTE"].format(
        chip_tool=self._chip_tool_path,
        node_id=node_id,
        endpoint_id=endpoint_id,
        cluster=cluster,
        attribute=attribute,
        value=value,
    )
    status_code = self._shell_with_regex(
        command, _REGEXES["WRITE_CLUSTER_ATTRIBUTE_RESPONSE"], raise_error=True)
    if int(status_code, 0) != 0:
      raise errors.DeviceError(
          f"{self._device_name} '{command}' responded with a non-zero "
          f"status code: {status_code}")

  @decorators.CapabilityLogDecorator(logger)
  def send(self, node_id: int, endpoint_id: int, cluster: str, command: str,
           arguments: Sequence[Any]) -> None:
    """Sends a command to a device with the given node id and endpoint.

    Args:
      node_id: Node ID assigned to the commissioned end device.
      endpoint_id: Endpoint ID within the node to read attribute from.
      cluster: Name of the cluster to send the command to (e.g. onoff).
      command: Name of the command to send (e.g. toggle).
      arguments: Command arguments.
    """
    command = _COMMANDS["SEND_CLUSTER_COMMMAND"].format(
        chip_tool=self._chip_tool_path,
        node_id=node_id,
        endpoint_id=endpoint_id,
        cluster=cluster,
        command=command,
        arguments=" ".join(map(str, arguments)),
    )
    status_code = self._shell_with_regex(
        command,
        _REGEXES["SEND_CLUSTER_COMMAND_RESPONSE"],
        raise_error=True,
        timeout=_TIMEOUTS["SEND_CLUSTER_COMMAND"])
    if int(status_code, 0) != 0:
      raise errors.DeviceError(
          f"{self._device_name} '{command}' responded with a non-zero "
          f"status code: {status_code}")

  @decorators.CapabilityLogDecorator(logger)
  def upgrade(self, build_file: str, build_id: str) -> None:
    """Installs chip-tool binary to the controller device.

    Args:
      build_file: Path to chip-tool binary on the host machine.
      build_id: Commit SHA the chip-tool binary is built at.
    """
    self._send_file_to_device(build_file, self._chip_tool_path)
    self._shell(
        _COMMANDS["WRITE_CHIP_TOOL_VERSION"].format(chip_tool_version=build_id))



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Docker flavor of file transfer capability (for devices using PTY transports)."""
import os

from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.capabilities.interfaces import file_transfer_base
from gazoo_device.utility import host_utils

logger = gdm_logger.get_logger()


class FileTransferDocker(file_transfer_base.FileTransferBase):
  """Docker-based file transfer capability."""

  def __init__(self, docker_container, device_name):
    """Initialize docker file transfer capability.

    Args:
        docker_container (str): name or id of the device docker container
          using the capability.
        device_name (str): name of the device using the capability
    """

    super().__init__(device_name=device_name)
    self._docker_container = docker_container

  @decorators.CapabilityLogDecorator(logger)
  def send_file_to_device(self, src, dest):
    """Copy a file from the local host to the device.

    Args:
        src (str): Absolute path of the source file on the host.
        dest (str): Absolute path of the destination on the device

    Raises:
        DeviceError: if the file transfer fails for any reason,
           or if the specified mode is invalid.

    Note:
       dest may be either a directory or a file name.
    """
    if not os.path.exists(src):
      raise errors.DeviceError("Device {} send to device failed. "
                               "Source file {} doesn't appear to exist.".format(
                                   self._device_name, src))

    logger.info("%s sending file to device. Source: %s Destination: %s",
                self._device_name, src, dest)
    host_utils.docker_cp_to_device(
        docker_container=self._docker_container,
        local_file_path=src,
        container_file_path=dest)

  @decorators.CapabilityLogDecorator(logger)
  def recv_file_from_device(self, src, dest="./"):
    """Copy a file from the device to the local host.

    Args:
        src (str): Absolute path of the source file on the device.
        dest (str): Absolute path of the destination on the host

    Raises:
      DeviceError: if the file transfer fails for any reason.
      RuntimeError: if the final file does not exist

    Note:
     dest can be directory or a file name.
    """
    dest_dir = os.path.dirname(dest)
    if dest_dir != "." and not os.path.exists(dest_dir):
      raise errors.DeviceError(
          "Device {} receive from device failed. "
          "Destination directory {} doesn't appear to exist.".format(
              self._device_name, dest_dir))

    logger.info("%s receiving file from device. Source: %s Destination: %s",
                self._device_name, src, dest)
    host_utils.docker_cp_from_device(
        docker_container=self._docker_container,
        local_file_path=dest,
        container_file_path=src)
    if not os.path.exists(dest):
      raise RuntimeError("Final file {} does not exist".format(dest))



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Default implementation of the PwRPC (Pigweed RPC) wifi capability."""

from typing import Any, Callable, Optional

from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.capabilities.interfaces import pwrpc_wifi_base
from gazoo_device.protos import wifi_service_pb2
from gazoo_device.switchboard.transports import pigweed_rpc_transport


_LOGGER = gdm_logger.get_logger()
_WIFI_RPC = "WiFi"
_ASCII_NULL_BYTES = b"\x00"


class PwRPCWifiDefault(pwrpc_wifi_base.PwRPCWifiBase):
  """Pigweed RPC wifi capability."""

  def __init__(self,
               device_name: str,
               switchboard_call: Callable[..., Any]):
    """Initializes an instance of the PwRPCWifiDefault capability.

    Args:
      device_name: Device name used for logging.
      switchboard_call: The switchboard.call method.
    """
    super().__init__(device_name=device_name)
    self._switchboard_call = switchboard_call

  def _verify_rpc_ack(self, ack: bool) -> None:
    """Helper method to verify rpc was acknowledged by device successfully.

    Args:
      ack: RPC ack.ok() value.

    Raises:
      DeviceError if the device failed to acknowledge the RPC.
    """
    if not ack:
      raise errors.DeviceError(f"{self._device_name} did not acknowledge "
                               "the RPC.")

  @decorators.DynamicProperty
  def channel(self) -> int:
    """Wifi channel used with current ssid connection."""
    ack, channel_bytes = self._switchboard_call(
        method=pigweed_rpc_transport.PigweedRPCTransport.rpc,
        method_args=(_WIFI_RPC, "GetChannel"),
        method_kwargs={})

    self._verify_rpc_ack(ack)
    return wifi_service_pb2.Channel.FromString(channel_bytes).channel

  @decorators.DynamicProperty
  def ssid(self) -> str:
    """Name of the SSID to which device is connected."""
    ack, ssid_bytes = self._switchboard_call(
        method=pigweed_rpc_transport.PigweedRPCTransport.rpc,
        method_args=(_WIFI_RPC, "GetSsid"),
        method_kwargs={})

    self._verify_rpc_ack(ack)
    ssid = wifi_service_pb2.Ssid.FromString(ssid_bytes).ssid
    # b/212183946: Remove any null characters from ssid name.
    return ssid.rstrip(_ASCII_NULL_BYTES).decode()

  @decorators.DynamicProperty
  def mac_address(self) -> str:
    """MAC address for the wifi interface.

    Returns:
      String representation of mac_address in lower case hexadecimal digits
      separated by colon (:). e.g. 24:0a:c4:f8:6c:d4
    """
    ack, mac_address_bytes = self._switchboard_call(
        method=pigweed_rpc_transport.PigweedRPCTransport.rpc,
        method_args=(_WIFI_RPC, "GetMacAddress"),
        method_kwargs={})

    self._verify_rpc_ack(ack)
    return wifi_service_pb2.MacAddress.FromString(mac_address_bytes).mac_address

  @decorators.DynamicProperty
  def wifi_interface(self) -> str:
    """Name of the interface used for wifi connection."""
    ack, wifi_interface_bytes = self._switchboard_call(
        method=pigweed_rpc_transport.PigweedRPCTransport.rpc,
        method_args=(_WIFI_RPC, "GetWiFiInterface"),
        method_kwargs={})

    self._verify_rpc_ack(ack)
    return wifi_service_pb2.WiFiInterface.FromString(
        wifi_interface_bytes).interface

  @decorators.DynamicProperty
  def ipv4_address(self) -> str:
    """IPv4 address for wifi interface."""
    ack, ip4_address_bytes = self._switchboard_call(
        method=pigweed_rpc_transport.PigweedRPCTransport.rpc,
        method_args=(_WIFI_RPC, "GetIP4Address"),
        method_kwargs={})

    self._verify_rpc_ack(ack)
    return wifi_service_pb2.IP4Address.FromString(ip4_address_bytes).address

  @decorators.CapabilityLogDecorator(_LOGGER)
  def connect(self,
              ssid: str,
              security_type: str,
              secret: Optional[str] = None) -> None:
    """Establishes wifi conection with given ssid.

    Args:
      ssid: Name of the wifi ssid.
      security_type: Wifi encryption protocol as defined in
        wifi_service_pb2.WIFI_SECURITY_TYPE.
      secret: Wifi passphrase.

    Raises:
      DeviceError with appropriate error code if connection failed.
    """
    connect_kwargs = {
        "ssid":
            ssid.encode(),
        "security_type":
            wifi_service_pb2.WIFI_SECURITY_TYPE.Value(security_type)
    }
    if secret is not None:
      connect_kwargs.update({"secret": secret.encode()})
    ack, connection_result_bytes = self._switchboard_call(
        method=pigweed_rpc_transport.PigweedRPCTransport.rpc,
        method_args=(_WIFI_RPC, "Connect"),
        method_kwargs=connect_kwargs)

    self._verify_rpc_ack(ack)
    connection_result = wifi_service_pb2.ConnectionResult.FromString(
        connection_result_bytes)
    result = connection_result.error
    if result != wifi_service_pb2.CONNECTION_ERROR.OK:
      raise errors.DeviceError(f"{self._device_name} failed to connect to wifi "
                               f"with {result}")

  @decorators.CapabilityLogDecorator(_LOGGER)
  def disconnect(self) -> None:
    """Disconnects from current wifi connection."""
    ack, _ = self._switchboard_call(
        method=pigweed_rpc_transport.PigweedRPCTransport.rpc,
        method_args=(_WIFI_RPC, "Disconnect"),
        method_kwargs={})
    self._verify_rpc_ack(ack)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Matter device base class for all vendor platforms."""
import os
from typing import Callable, Dict, List, NoReturn, Optional, Tuple

from gazoo_device import console_config
from gazoo_device import custom_types
from gazoo_device import decorators
from gazoo_device import gdm_logger
from gazoo_device.base_classes import gazoo_device_base
from gazoo_device.capabilities import device_power_default
from gazoo_device.capabilities import matter_endpoints_accessor
from gazoo_device.capabilities import pwrpc_button_default
from gazoo_device.capabilities import pwrpc_common_default
from gazoo_device.capabilities.matter_endpoints import color_temperature_light
from gazoo_device.capabilities.matter_endpoints import dimmable_light
from gazoo_device.capabilities.matter_endpoints import door_lock
from gazoo_device.capabilities.matter_endpoints import on_off_light
from gazoo_device.capabilities.matter_endpoints import temperature_sensor
from gazoo_device.protos import attributes_service_pb2
from gazoo_device.protos import button_service_pb2
from gazoo_device.protos import descriptor_service_pb2
from gazoo_device.protos import device_service_pb2
from gazoo_device.protos import wifi_service_pb2
from gazoo_device.utility import usb_utils


logger = gdm_logger.get_logger()
BAUDRATE = 115200
_RPC_TIMEOUT = 5  # seconds
_DEFAULT_BOOTUP_TIMEOUT_SECONDS = 30
_CONNECTION_TIMEOUT_SECONDS = 10
_REBOOT_METHODS = ("pw_rpc", "soft", "hard")


class MatterDeviceBase(gazoo_device_base.GazooDeviceBase):
  """Matter device base class."""
  COMMUNICATION_TYPE = "PigweedSerialComms"
  _OWNER_EMAIL = "gdm-authors@google.com"
  _COMMUNICATION_KWARGS = {"protobufs": (attributes_service_pb2,
                                         button_service_pb2,
                                         descriptor_service_pb2,
                                         device_service_pb2,
                                         wifi_service_pb2),
                           "baudrate": BAUDRATE}
  # Should be overridden in the derived platform classes which support button
  # RPCs.
  VALID_BUTTON_IDS = ()

  def __init__(self,
               manager,
               device_config,
               log_file_name=None,
               log_directory=None):
    super().__init__(
        manager,
        device_config,
        log_file_name=log_file_name,
        log_directory=log_directory)
    self._timeouts.update({"CONNECTED": _CONNECTION_TIMEOUT_SECONDS})

  def get_console_configuration(self) -> console_config.ConsoleConfiguration:
    """Returns the interactive console configuration."""
    return console_config.get_log_only_configuration()

  @decorators.LogDecorator(logger)
  def get_detection_info(self) -> Tuple[Dict[str, str], Dict[str, str]]:
    """Gets the persistent and optional attributes of a device during setup.

    Returns:
      Dictionary of persistent attributes and dictionary of
      optional attributes.
    """
    persistent_dict = self.props["persistent_identifiers"]
    address = persistent_dict["console_port_name"]
    persistent_dict["serial_number"] = (
        usb_utils.get_serial_number_from_path(address))
    persistent_dict["model"] = "PROTO"
    persistent_dict.update(
        usb_utils.get_usb_hub_info(self.communication_address))
    return persistent_dict, {}

  @decorators.PersistentProperty
  def health_checks(self) -> List[Callable[[], None]]:
    """Returns list of methods to execute as health checks."""
    return [
        self.check_power_cycling_ready,
        self.check_power_on,
        self.check_device_connected,
        self.check_create_switchboard,
    ]

  @decorators.health_check
  def check_power_on(self) -> None:
    """Checks that the power is on."""
    if self.device_power.healthy:
      self.device_power.on()

  @classmethod
  def is_connected(cls,
                   device_config: custom_types.ManagerDeviceConfigDict) -> bool:
    """Returns True if the device is connected to the host."""
    return os.path.exists(device_config["persistent"]["console_port_name"])

  @decorators.DynamicProperty
  def firmware_version(self) -> str:
    """Firmware version of the device."""
    return str(self.pw_rpc_common.software_version)

  @decorators.DynamicProperty
  def vendor_id(self) -> int:
    """Vendor ID of the device."""
    return self.pw_rpc_common.vendor_id

  @decorators.DynamicProperty
  def product_id(self) -> int:
    """Product ID of the device."""
    return self.pw_rpc_common.product_id

  @decorators.PersistentProperty
  def device_usb_hub_name(self) -> Optional[str]:
    """The name of the USB hub for the device or None if not configured."""
    return self.props["persistent_identifiers"].get("device_usb_hub_name", None)

  @decorators.PersistentProperty
  def device_usb_port(self) -> Optional[str]:
    """The port number on the USB hub or None if not configured."""
    return self.props["persistent_identifiers"].get("device_usb_port", None)

  @decorators.DynamicProperty
  def pairing_code(self) -> int:
    """Pairing code of the device."""
    return self.pw_rpc_common.pairing_info.code

  @decorators.DynamicProperty
  def pairing_discriminator(self) -> int:
    """Pairing discriminator of the device."""
    return self.pw_rpc_common.pairing_info.discriminator

  @decorators.DynamicProperty
  def pairing_state(self) -> bool:
    """Pairing state of the device."""
    return self.pw_rpc_common.pairing_state

  @decorators.DynamicProperty
  def qr_code(self) -> int:
    """QR code of the device."""
    return self.pw_rpc_common.qr_code

  @decorators.DynamicProperty
  def qr_code_url(self) -> int:
    """QR code URL of the device."""
    return self.pw_rpc_common.qr_code_url

  @decorators.LogDecorator(logger)
  def reboot(self, no_wait: bool = False, method: str = "pw_rpc") -> None:
    """Reboots the device.

    Args:
      no_wait: Return before reboot completes.
      method: Reboot technique to use. Valid methods
       ["pw_rpc", "soft", "hard"].

    Raises:
      ValueError: If invalid reboot method is specified.
    """
    if method not in _REBOOT_METHODS:
      raise ValueError(
          f"Method {method} not recognized. Supported methods: "
          f"{_REBOOT_METHODS}"
      )
    if method == "hard":
      self.device_power.cycle(no_wait=no_wait)
    else:
      self.pw_rpc_common.reboot(verify=not no_wait)

  @decorators.LogDecorator(logger)
  def factory_reset(self, no_wait: bool = False) -> None:
    """Factory resets the device.

    Args:
      no_wait: Return before reboot completes.
    """
    self.pw_rpc_common.factory_reset(verify=not no_wait)

  @decorators.LogDecorator(logger)
  def shell(self,
            command: str,
            command_name: str = "shell",
            timeout: Optional[int] = None,
            port: int = 0,
            searchwindowsize: int = 2000,
            include_return_code: bool = False) -> NoReturn:
    """Sends command and returns response.

    Args:
      command: Command to send to the device.
      command_name: Optional identifier to use in logs for this command.
      timeout: Seconds to wait for pattern after command sent.
        If None, the default shell timeout is used.
      port: Which port to send on, 0 or 1. Default: 0.
      searchwindowsize: Number of the last bytes to look at
      include_return_code: Flag indicating return code should be returned

    Raises:
      NotImplementedError:
      shell method is not implemented for Matter device.
    """
    raise NotImplementedError("shell not implemented for Matter device.")

  @decorators.LogDecorator(logger)
  def wait_for_bootup_complete(
      self, timeout: int = _DEFAULT_BOOTUP_TIMEOUT_SECONDS) -> None:
    """Wait until the device finishes booting up and is ready for testing.

    Args:
      timeout: Max time to wait for the device to finish booting up.
    """
    self.pw_rpc_common.wait_for_bootup_complete(timeout)

  @decorators.CapabilityDecorator(pwrpc_button_default.PwRPCButtonDefault)
  def pw_rpc_button(self):
    """PwRPCButtonDefault capability to send RPC command."""
    return self.lazy_init(
        pwrpc_button_default.PwRPCButtonDefault,
        device_name=self.name,
        valid_button_ids=self.VALID_BUTTON_IDS,
        switchboard_call=self.switchboard.call,
        rpc_timeout_s=_RPC_TIMEOUT)

  @decorators.CapabilityDecorator(pwrpc_common_default.PwRPCCommonDefault)
  def pw_rpc_common(self) -> pwrpc_common_default.PwRPCCommonDefault:
    """PwRPCCommonDefault capability to send RPC command."""
    return self.lazy_init(
        pwrpc_common_default.PwRPCCommonDefault,
        device_name=self.name,
        switchboard_call=self.switchboard.call,
        rpc_timeout_s=_RPC_TIMEOUT)

  @decorators.CapabilityDecorator(
      matter_endpoints_accessor.MatterEndpointsAccessor)
  def matter_endpoints(
      self) -> matter_endpoints_accessor.MatterEndpointsAccessor:
    """Generic Matter endpoint instance."""
    return self.lazy_init(
        matter_endpoints_accessor.MatterEndpointsAccessor,
        device_name=self.name,
        switchboard_call=self.switchboard.call,
        rpc_timeout_s=_RPC_TIMEOUT
    )

  @decorators.CapabilityDecorator(device_power_default.DevicePowerDefault)
  def device_power(self) -> device_power_default.DevicePowerDefault:
    """Capability to manipulate device power through Cambrionix."""
    return self.lazy_init(
        device_power_default.DevicePowerDefault,
        device_name=self.name,
        create_device_func=self.get_manager().create_device,
        default_hub_type="cambrionix",
        props=self.props,
        usb_ports_discovered=True,
        switchboard_inst=self.switchboard,
        wait_for_bootup_complete_fn=self.wait_for_bootup_complete,
        wait_for_connection_fn=self.check_device_connected,
        usb_hub_name_prop="device_usb_hub_name",
        usb_port_prop="device_usb_port")

  # ******************** Matter endpoint aliases ******************** #
  @decorators.CapabilityDecorator(
      color_temperature_light.ColorTemperatureLightEndpoint)
  def color_temperature_light(
      self) -> color_temperature_light.ColorTemperatureLightEndpoint:
    """Matter Color Temperature Light endpoint instance.

    Returns:
      Color Temperature Light endpoint instance.

    Raises:
      DeviceError when Color Temperate Light endpoint is not supported on the
      device.
    """
    return self.matter_endpoints.get_endpoint_instance_by_class(
        color_temperature_light.ColorTemperatureLightEndpoint)

  @decorators.CapabilityDecorator(dimmable_light.DimmableLightEndpoint)
  def dimmable_light(self) -> dimmable_light.DimmableLightEndpoint:
    """Matter Dimmable Light endpoint instance.

    Returns:
      Dimmable Light endpoint instance.

    Raises:
      DeviceError when Dimmable Light endpoint is not supported on the
      device.
    """
    return self.matter_endpoints.get_endpoint_instance_by_class(
        dimmable_light.DimmableLightEndpoint)

  @decorators.CapabilityDecorator(door_lock.DoorLockEndpoint)
  def door_lock(self) -> door_lock.DoorLockEndpoint:
    """Matter Door Lock endpoint instance.

    Returns:
      Door Lock endpoint instance.

    Raises:
      DeviceError when Door Lock endpoint is not supported on the
      device.
    """
    return self.matter_endpoints.get_endpoint_instance_by_class(
        door_lock.DoorLockEndpoint)

  @decorators.CapabilityDecorator(on_off_light.OnOffLightEndpoint)
  def on_off_light(self) -> on_off_light.OnOffLightEndpoint:
    """Matter OnOff Light endpoint instance.

    Returns:
      OnOff Light endpoint instance.

    Raises:
      DeviceError when OnOff Light endpoint is not supported on the
      device.
    """
    return self.matter_endpoints.get_endpoint_instance_by_class(
        on_off_light.OnOffLightEndpoint)

  @decorators.CapabilityDecorator(temperature_sensor.TemperatureSensorEndpoint)
  def temperature_sensor(self) -> temperature_sensor.TemperatureSensorEndpoint:
    """Matter Temperature Sensor endpoint instance.

    Returns:
      Temperature Sensor endpoint instance.

    Raises:
      DeviceError when Temperature Sensor endpoint is not supported on the
      device.
    """
    return self.matter_endpoints.get_endpoint_instance_by_class(
        temperature_sensor.TemperatureSensorEndpoint)
  # ***************************************************************** #



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for analyzing GDM code at runtime (reflection)."""
import importlib
import inspect
import os
import types
from typing import Any, Callable, Collection, Dict, List, Optional, Set, Type

from gazoo_device import config
from gazoo_device import gdm_logger

logger = gdm_logger.get_logger()

_PACKAGE_DIR = "gazoo_device/"


def get_all_subclasses_in_module(
    parent_class: Type[Any],
    module: types.ModuleType,
    exclude_private: bool = True,
    exclude_abstract: bool = False) -> List[Type[Any]]:
  """Returns all classes derived from parent_class in the module.

  Args:
    parent_class: Class object from which the subclasses must be derived.
    module: Module to scan for subclasses.
    exclude_private: If True, do not include private classes.
    exclude_abstract: If True, do not include abstract classes.

  Returns:
    Subclasses of parent_class defined in the module.
  """
  subclasses = []
  for name, member in inspect.getmembers(module):
    if (inspect.isclass(member)
        and (not exclude_abstract or not inspect.isabstract(member))
        and (not exclude_private or not name.startswith("_"))
        and issubclass(member, parent_class)):
      subclasses.append(member)
  return subclasses


def _get_module_path(module: types.ModuleType) -> str:
  """Returns the module path."""
  return module.__file__


def get_all_subclasses_in_package(
    parent_class: Type[Any],
    package: types.ModuleType,
    excluded_modules: Optional[Collection[types.ModuleType]] = None,
    exclude_private: bool = True,
    exclude_abstract: bool = False,
    module_path_getter: Callable[[types.ModuleType], str] = _get_module_path
) -> Dict[types.ModuleType, Set[Type[Any]]]:
  """Returns classes derived from parent_class in modules within the package.

  Does not recurse into subpackages (subdirectories are not scanned).

  Args:
    parent_class: Class object from which the subclasses must be derived.
    package: __init__.py module of the package to scan.
    excluded_modules: Modules to exclude from the scan.
    exclude_private: If True, do not include private classes.
    exclude_abstract: If True, do not include abstract classes.
    module_path_getter: Function to retrieve the file path of a module.

  Returns:
    Mapping from module to set of classes derived from parent_class defined in
    that module.
  """
  if excluded_modules is None:
    excluded_modules = set()

  modules = [
      module for module in _list_package_modules(package, module_path_getter)
      if module not in excluded_modules
  ]

  module_to_classes = {}
  for module in modules:
    subclasses = get_all_subclasses_in_module(parent_class=parent_class,
                                              module=module,
                                              exclude_private=exclude_private,
                                              exclude_abstract=exclude_abstract)
    if subclasses:
      module_to_classes[module] = set(subclasses)
  return module_to_classes


def _list_package_modules(
    package: types.ModuleType,
    module_path_getter: Callable[[types.ModuleType], str] = _get_module_path
) -> List[types.ModuleType]:
  """Returns a list of all modules defined in the package.

  Subpackages (subdirectories) are not scanned.

  Args:
    package: __init__.py module of the package to scan. For example,
      gazoo_device.capabilities.
    module_path_getter: Function to retrieve the file path of a module.
  """
  suffix = ".py"
  suffix_len = len(suffix)

  package_path = module_path_getter(package)
  package_dir = os.path.dirname(package_path)
  files_in_package = os.listdir(package_dir)

  module_names = [
      a_file[:-suffix_len]
      for a_file in files_in_package
      if a_file.endswith(suffix) and a_file != "__init__.py"
  ]

  modules = []
  for module_name in module_names:
    modules.append(
        importlib.import_module(
            ".{}".format(module_name), package=package.__name__))
  return modules



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Interface for the file transfer capability."""
import abc
from gazoo_device.capabilities.interfaces import capability_base


class FileTransferBase(capability_base.CapabilityBase):
  """Abstract Base class for the file transfer capability."""

  @abc.abstractmethod
  def send_file_to_device(self, src, dest):
    """Copy a file from the local host to the device.

    Args:
        src (str): Absolute path of the source file on the host.
        dest (str): Absolute path of the destination on the device

    Raises:
        DeviceError: if the file transfer fails for any reason,
          or if the specified mode is invalid.

    Note:
        dest may be either a directory or a file name.
    """

  @abc.abstractmethod
  def recv_file_from_device(self, src, dest):
    """Copy a file from the device to the local host.

    Args:
        src (str): Absolute path of the source file on the device.
        dest (str): Absolute path of the destination on the host

    Raises:
        DeviceError: if the file transfer fails for any reason.

    Note:
        dest can be directory or a file name.
    """
